"""Module containing different I/O functions to load time log data, subject condition lists, questionnaire data, etc."""

import datetime
from pathlib import Path
from typing import Dict, Literal, Optional, Sequence, Tuple, Union

import numpy as np
import pandas as pd
import pytz
from biopsykit.utils._datatype_validation_helper import _assert_file_extension, _assert_has_columns, _assert_is_dtype
from biopsykit.utils._types import path_t
from biopsykit.utils.dataframe_handling import convert_nan
from biopsykit.utils.datatype_helper import (
    CodebookDataFrame,
    SubjectConditionDataFrame,
    SubjectConditionDict,
    _CodebookDataFrame,
    _SubjectConditionDataFrame,
    is_codebook_dataframe,
    is_subject_condition_dataframe,
    is_subject_condition_dict,
)
from biopsykit.utils.exceptions import ValidationError
from biopsykit.utils.file_handling import is_excel_file
from biopsykit.utils.time import tz
from nilspodlib import Dataset

__all__ = [
    "load_long_format_csv",
    "load_time_log",
    "load_subject_condition_list",
    "load_questionnaire_data",
    "load_pandas_dict_excel",
    "load_codebook",
    "convert_time_log_datetime",
    "write_pandas_dict_excel",
    "write_result_dict",
]


def load_long_format_csv(file_path: path_t, index_cols: Optional[Union[str, Sequence[str]]] = None) -> pd.DataFrame:
    """Load dataframe stored as long-format from file.

    Parameters
    ----------
    file_path : :class:`~pathlib.Path` or str
        path to file. Must be a csv file
    index_cols : str or list of str, optional
        column name (or list of such) of index columns to be used as MultiIndex in the resulting long-format
        dataframe or ``None`` to use all columns except the last one as index columns.
        Default: ``None``

    Returns
    -------
    :class:`~pandas.DataFrame`
        dataframe in long-format

    """
    # ensure pathlib
    file_path = Path(file_path)
    _assert_file_extension(file_path, expected_extension=[".csv"])

    data = pd.read_csv(file_path)

    if index_cols is None:
        index_cols = list(data.columns)[:-1]
    _assert_has_columns(data, [index_cols])

    return data.set_index(index_cols)


def load_time_log(
    file_path: path_t,
    subject_col: Optional[str] = None,
    condition_col: Optional[str] = None,
    additional_index_cols: Optional[Union[str, Sequence[str]]] = None,
    phase_cols: Optional[Union[Sequence[str], Dict[str, str]]] = None,
    continuous_time: Optional[bool] = True,
    **kwargs,
) -> pd.DataFrame:
    """Load time log information from file.

    This function can be used to load a file containing "time logs", i.e.,
    information about start and stop times of recordings or recording phases per subject.

    Parameters
    ----------
    file_path : :class:`~pathlib.Path` or str
        path to time log file. Must either be an Excel or csv file
    subject_col : str, optional
        name of column containing subject IDs or ``None`` to use default column name ``subject``.
        According to BioPsyKit's convention, the subject ID column is expected to have the name ``subject``.
        If the subject ID column in the file has another name, the column will be renamed in the dataframe
        returned by this function.
    condition_col : str, optional
        name of column containing condition assignments or ``None`` to use default column name ``condition``.
        According to BioPsyKit's convention, the condition column is expected to have the name ``condition``.
        If the condition column in the file has another name, the column will be renamed in the dataframe
        returned by this function.
    additional_index_cols : str, list of str, optional
        additional index levels to be added to the dataframe.
        Can either be a string or a list strings to indicate column name(s) that should be used as index level(s),
        or ``None`` for no additional index levels. Default: ``None``
    phase_cols : list of str or dict, optional
        list of column names that contain time log information or ``None`` to use all columns.
        If the column names of the time log dataframe should have different names than the columns in the file,
        a dict specifying the mapping (column_name : new_column_name) can be passed. Default: ``None``
    continuous_time: bool, optional
        flag indicating whether phases are continuous, i.e., whether the end of the previous phase is also the
        beginning of the next phase or not. Default: ``True``.
        If ``continuous_time`` is set to ``False``, the start and end columns of all phases must have the
        suffixes "_start" and "_end", respectively
    **kwargs
        Additional parameters that are passed to :func:`pandas.read_csv` or :func:`pandas.read_excel`

    Returns
    -------
    :class:`~pandas.DataFrame`
        dataframe with time log information

    Raises
    ------
    :exc:`~biopsykit.utils.exceptions.FileExtensionError`
        if file format is none of [".xls", ".xlsx", ".csv"]
    :exc:`~biopsykit.utils.exceptions.ValidationError`
        if ``continuous_time`` is ``False``, but "start" and "end" time columns of each phase do not match or
        none of these columns were found in the dataframe

    Examples
    --------
    >>> import biopsykit as bp
    >>> file_path = "./timelog.csv"
    >>> # Example 1:
    >>> # load time log file into a pandas dataframe
    >>> data = bp.io.load_time_log(file_path)
    >>> # Example 2:
    >>> # load time log file into a pandas dataframe and specify the "ID" column
    >>> # (instead of the default "subject" column) in the time log file to be the index of the dataframe
    >>> data = bp.io.load_time_log(file_path, subject_col="ID")
    >>> # Example 3:
    >>> # load time log file into a pandas dataframe and specify the columns "Phase1", "Phase2", and "Phase3"
    >>> # to be used for extracting time information
    >>> data = bp.io.load_time_log(
    >>>     file_path, phase_cols=["Phase1", "Phase2", "Phase3"]
    >>> )
    >>> # Example 4:
    >>> # load time log file into a pandas dataframe and specify the column "ID" as subject column, the column "Group"
    >>> # as condition column, as well as the column "Time" as additional index column.
    >>> data = bp.io.load_time_log(file_path,
    >>>     subject_col="ID",
    >>>     condition_col="Group",
    >>>     additional_index_cols=["Time"],
    >>>     phase_cols=["Phase1", "Phase2", "Phase3"]
    >>> )

    """
    # ensure pathlib
    file_path = Path(file_path)

    _assert_file_extension(file_path, expected_extension=[".xls", ".xlsx", ".csv"])
    # assert times in the excel sheet are imported as strings,
    # not to be automatically converted into datetime objects
    kwargs["dtype"] = str
    data = _load_dataframe(file_path, **kwargs)

    data, index_cols = _sanitize_index_cols(data, subject_col, condition_col, additional_index_cols)
    data = _apply_index_cols(data, index_cols=index_cols)
    data = _apply_phase_cols(data, phase_cols=phase_cols)
    data.columns.name = "phase"

    if not continuous_time:
        data = _parse_time_log_not_continuous(data, index_cols)

    for val in data.to_numpy().flatten():
        if val is np.nan:
            continue
        _assert_is_dtype(val, str)

    return data


def convert_time_log_datetime(
    time_log: pd.DataFrame,
    dataset: Optional[Dataset] = None,
    df: Optional[pd.DataFrame] = None,
    date: Optional[Union[str, datetime.datetime]] = None,
    timezone: Optional[Union[str, datetime.tzinfo]] = None,
) -> pd.DataFrame:
    """Convert the time log information into datetime objects.

    This function converts time log information (containing only time, but no date)
    into datetime objects, thus, adds the `start date` of the recording. To specify the recording date,
    either a NilsPod :class:`~nilspodlib.dataset.Dataset` or a pandas dataframe with a :class:`~pandas.DatetimeIndex`
    must be supplied from which the recording date can be extracted.
    As an alternative, the date can be specified explicitly via ``date`` parameter.

    Parameters
    ----------
    time_log : :class:`~pandas.DataFrame`
        pandas dataframe with time log information
    dataset : :class:`~nilspodlib.dataset.Dataset`, optional
        NilsPod Dataset object extract time and date information. Default: ``None``
    df : :class:`~pandas.DataFrame`, optional
        dataframe with :class:`~pandas.DatetimeIndex` to extract time and date information. Default: ``None``
    date : str or datetime, optional
        datetime object or date string used to convert time log information into datetime.
        If ``date`` is a string, it must be supplied in a common date format, e.g. "dd.mm.yyyy" or "dd/mm/yyyy".
        Default: ``None``
    timezone : str or :class:`datetime.tzinfo`, optional
        timezone of the acquired data to convert, either as string of as tzinfo object.
        Default: "Europe/Berlin"

    Returns
    -------
    :class:`~pandas.DataFrame`
        pandas dataframe with log time converted into datetime

    Raises
    ------
    ValueError
        if none of ``dataset``, ``df`` and ``date`` are supplied as argument,
        or if index of ``df`` is not a :class:`~pandas.DatetimeIndex`

    """
    if dataset is None and date is None and df is None:
        raise ValueError("Either `dataset`, `df` or `date` must be supplied as argument to retrieve date information!")

    date = _extract_date(dataset, df, date)

    if timezone is None:
        timezone = tz
    if isinstance(timezone, str):
        timezone = pytz.timezone(timezone)

    if isinstance(time_log.to_numpy().flatten()[0], str):
        # convert time strings into datetime.time object
        time_log = time_log.applymap(pd.to_datetime)
        time_log = time_log.applymap(lambda val: val.time())

    time_log = time_log.applymap(lambda x: timezone.localize(datetime.datetime.combine(date, x)))
    return time_log


def load_atimelogger_file(file_path: path_t, timezone: Optional[Union[datetime.tzinfo, str]] = None) -> pd.DataFrame:
    """Load time log file exported from the aTimeLogger app.

    The resulting dataframe will have one row and start and end times of the single phases as columns.

    Parameters
    ----------
    file_path : :class:`~pathlib.Path` or str
        path to time log file. Must a csv file
    timezone : str or :class:`datetime.tzinfo`, optional
        timezone of the time logs, either as string or as `tzinfo` object.
        Default: 'Europe/Berlin'

    Returns
    -------
    :class:`~pandas.DataFrame`
        time log dataframe

    See Also
    --------
    :func:`~biopsykit.utils.io.convert_time_log_datetime`
        convert timelog dataframe into dictionary
    `aTimeLogger app <https://play.google.com/store/apps/details?id=com.aloggers.atimeloggerapp>`_

    """
    # ensure pathlib
    file_path = Path(file_path)
    _assert_file_extension(file_path, expected_extension=[".csv"])

    if timezone is None:
        timezone = tz

    timelog = pd.read_csv(file_path)
    # find out if file is german or english and get the right column names
    if "Aktivitätstyp" in timelog.columns:
        phase_col = "Aktivitätstyp"
        time_cols = ["Von", "Bis"]
    elif "Activity type" in timelog.columns:
        phase_col = "Activity type"
        time_cols = ["From", "To"]
    else:
        phase_col = "phase"
        time_cols = ["start", "end"]

    timelog = timelog.set_index(phase_col)
    timelog = timelog[time_cols]

    timelog = timelog.rename(columns={time_cols[0]: "start", time_cols[1]: "end"})
    timelog.index.name = "phase"
    timelog.columns.name = "start_end"

    timelog = timelog.apply(pd.to_datetime, axis=1).applymap(lambda val: val.tz_localize(timezone))
    timelog = pd.DataFrame(timelog.T.unstack(), columns=["time"])
    timelog = timelog[::-1].reindex(["start", "end"], level="start_end")
    timelog = timelog.T
    return timelog


def convert_time_log_dict(
    timelog: Union[pd.DataFrame, pd.Series], time_format: Optional[Literal["str", "time"]] = "time"
) -> Dict[str, Tuple[Union[str, datetime.time]]]:
    """Convert time log into dictionary.

    The resulting dictionary will have the phase names as keys and a tuple with start and end times as values.

    Parameters
    ----------
    timelog : :class:`~pandas.DataFrame` or :class:`~pandas.Series`
        dataframe or series containing timelog information
    time_format : "str" or "time", optional
        "str" to convert entries in dictionary to string, "time" to keep them as :class:`~datetime.time` objects.
        Default: "time"

    Returns
    -------
    dict
        dictionary with start and end times of each phase

    See Also
    --------
    :func:`biopsykit.utils.data_processing.split_data`
        split data based on time intervals

    """
    timelog = timelog.T.unstack()["time"]
    # assert correct order
    timelog = timelog[["start", "end"]]
    timelog_dict = timelog.to_dict(orient="index")
    timelog_dict = {key: tuple(v.time() for v in val.values()) for key, val in timelog_dict.items()}
    if time_format == "str":
        timelog_dict = {key: tuple(map(str, val)) for key, val in timelog_dict.items()}
    return timelog_dict


def load_subject_condition_list(
    file_path: path_t,
    subject_col: Optional[str] = None,
    condition_col: Optional[str] = None,
    return_dict: Optional[bool] = False,
    **kwargs,
) -> Union[SubjectConditionDataFrame, SubjectConditionDict]:
    """Load subject condition assignment from file.

    This function can be used to load a file that contains the assignment of subject IDs to study conditions.
    It will return a dataframe or a dictionary that complies with BioPsyKit's naming convention, i.e.,
    the subject ID index will be named ``subject`` and the condition column will be named ``condition``.

    Parameters
    ----------
    file_path : :class:`~pathlib.Path` or str
        path to time log file. Must either be an Excel or csv file
    subject_col : str, optional
        name of column containing subject IDs or ``None`` to use default column name ``subject``.
        According to BioPsyKit's convention, the subject ID column is expected to have the name ``subject``.
        If the subject ID column in the file has another name, the column will be renamed in the dataframe
        returned by this function.
    condition_col : str, optional
        name of column containing condition assignments or ``None`` to use default column name ``condition``.
        According to BioPsyKit's convention, the condition column is expected to have the name ``condition``.
        If the condition column in the file has another name, the column will be renamed in the dataframe
        returned by this function.
    return_dict : bool, optional
        whether to return a dict with subject IDs per condition (``True``) or a dataframe (``False``).
        Default: ``False``
    **kwargs
        Additional parameters that are passed tos :func:`pandas.read_csv` or :func:`pandas.read_excel`

    Returns
    -------
    :class:`~biopsykit.utils.datatype_helper.SubjectConditionDataFrame` or
    :class:`~biopsykit.utils.datatype_helper.SubjectConditionDict`
        a standardized pandas dataframe with subject IDs and condition assignments (if ``return_dict`` is ``False``) or
        a standardized dict with subject IDs per group (if ``return_dict`` is ``True``)

    Raises
    ------
    :exc:`~biopsykit.utils.exceptions.FileExtensionError`
        if file is not a csv or Excel file
    :exc:`~biopsykit.utils.exceptions.ValidationError`
        if result is not a :class:`~biopsykit.utils.datatype_helper.SubjectConditionDataFrame` or a
        :class:`~biopsykit.utils.datatype_helper.SubjectConditionDict`

    """
    # ensure pathlib
    file_path = Path(file_path)
    _assert_file_extension(file_path, expected_extension=[".xls", ".xlsx", ".csv"])

    data = _load_dataframe(file_path, **kwargs)

    if subject_col is None:
        subject_col = "subject"

    if condition_col is None:
        condition_col = "condition"

    _assert_has_columns(data, [[subject_col, condition_col]])

    if subject_col != "subject":
        # rename column
        subject_col = {subject_col: "subject"}
        data = data.rename(columns=subject_col)
        subject_col = "subject"

    if condition_col != "condition":
        # rename column
        condition_col = {condition_col: "condition"}
        data = data.rename(columns=condition_col)
        condition_col = "condition"
    data = data.set_index(subject_col)

    if return_dict:
        data = data.groupby(condition_col).groups
        is_subject_condition_dict(data)
        return data
    is_subject_condition_dataframe(data)
    return _SubjectConditionDataFrame(data)


def load_questionnaire_data(
    file_path: path_t,
    subject_col: Optional[str] = None,
    condition_col: Optional[str] = None,
    additional_index_cols: Optional[Union[str, Sequence[str]]] = None,
    replace_missing_vals: Optional[bool] = True,
    remove_nan_rows: Optional[bool] = True,
    sheet_name: Optional[Union[str, int]] = 0,
    **kwargs,
) -> pd.DataFrame:
    """Load questionnaire data from file.

    The resulting dataframe will comply with BioPsyKit's naming conventions, i.e., the subject ID index will be
    named ``subject`` and a potential condition index will be named ``condition``.

    Parameters
    ----------
    file_path : :class:`~pathlib.Path` or str
        path to time log file. Must either be an Excel or csv file
    subject_col : str, optional
        name of column containing subject IDs or ``None`` to use default column name ``subject``.
        According to BioPsyKit's convention, the subject ID column is expected to have the name ``subject``.
        If the subject ID column in the file has another name, the column will be renamed in the dataframe
        returned by this function.
    condition_col : str, optional
        name of column containing condition assignments or ``None`` to use default column name ``condition``.
        According to BioPsyKit's convention, the condition column is expected to have the name ``condition``.
        If the condition column in the file has another name, the column will be renamed in the dataframe
        returned by this function.
    additional_index_cols : str, list of str, optional
        additional index levels to be added to the dataframe.
        Can either be a string or a list strings to indicate column name(s) that should be used as index level(s),
        or ``None`` for no additional index levels. Default: ``None``
    replace_missing_vals : bool, optional
        ``True`` to replace encoded "missing values" from software like SPSS (e.g. -77, -99, or -66)
        to "actual" missing values (NaN).
        Default: ``True``
    remove_nan_rows : bool, optional
        ``True`` to remove rows that only contain NaN values (except the index cols), ``False`` to keep NaN rows.
        Default: ``True``
    sheet_name : str or int, optional
        sheet_name identifier (str) or sheet_name index (int) if file is an Excel file.
        Default: 0 (i.e. first sheet in Excel file)

    Returns
    -------
    :class:`~pandas.DataFrame`
        dataframe with imported questionnaire data

    Raises
    ------
    :class:`~biopsykit.utils.exceptions.FileExtensionError`
        if file format is none of [".xls", ".xlsx", ".csv"]

    """
    # ensure pathlib
    file_path = Path(file_path)

    _assert_file_extension(file_path, expected_extension=[".xls", ".xlsx", ".csv"])
    if file_path.suffix != ".csv":
        kwargs["sheet_name"] = sheet_name
    data = _load_dataframe(file_path, **kwargs)
    data, index_cols = _sanitize_index_cols(data, subject_col, condition_col, additional_index_cols)
    data = _apply_index_cols(data, index_cols=index_cols)

    if replace_missing_vals:
        data = convert_nan(data)
    if remove_nan_rows:
        data = data.dropna(how="all")
    return data


def load_codebook(file_path: path_t, **kwargs) -> CodebookDataFrame:
    """Load codebook from file.

    A codebook is used to convert numerical values from a dataframe (e.g., from questionnaire data)
    to categorical values.


    Parameters
    ----------
    file_path : :class:`~pathlib.Path` or str
        file path to codebook
    **kwargs
        additional arguments to pass to :func:`pandas.read_csv` or :func:`pandas.read_excel`


    Returns
    -------
    :class:`~pandas.DataFrame`
        :obj:`~biopsykit.utils.datatype_helper.CodebookDataFrame`, a dataframe in a standardized format


    See Also
    --------
    :func:`~biopsykit.utils.dataframe_handling.apply_codebook`
        apply codebook to data

    """
    # ensure pathlib
    file_path = Path(file_path)

    _assert_file_extension(file_path, expected_extension=[".xls", ".xlsx", ".csv"])
    if file_path.suffix in [".xls", ".xlsx"]:
        data = pd.read_excel(file_path, **kwargs)
    else:
        data = pd.read_csv(file_path, **kwargs)

    _assert_has_columns(data, [["variable"]])
    data = data.set_index("variable")
    data.columns = data.columns.astype(int)
    is_codebook_dataframe(data)

    return _CodebookDataFrame(data)


# def load_stroop_inquisit_data(folder_path: path_t, cols: Optional[Sequence[str]] = None) -> Dict[str, pd.DataFrame]:
#     """Load Inquisit data collected during "Stroop Test".
#
#     Stroop Test reusults (mean response time, number of correct answers, etc.)
#     are exported per Stroop phase and are stored in a common folder. This function loads all exported `.iqdat` files,
#     transforms them into dataframes and combines them into a dictionary.
#
#     Parameters
#     ----------
#     folder_path : :any:`pathlib.Path` or str
#         path to the folder in which the Stroop test export files are stored
#     cols : list of str, optional
#         names of columns which should be imported and added to the dictionary
#
#     Returns
#     -------
#     dict
#         dictionary with Stroop Test parameters per Stroop Phase
#
#     """
#     dict_stroop = {}
#     # ensure pathlib
#     folder_path = Path(folder_path)
#     # look for all Inquisit files in the folder
#     dataset_list = list(sorted(folder_path.glob("*.iqdat")))
#     subject = ""
#     # iterate through data
#     for data_path in dataset_list:
#         df_stroop = pd.read_csv(data_path, sep="\t")
#         if subject != df_stroop["subject"][0]:
#             dict_stroop = {}
#         # set subject, stroop phase
#         subject = df_stroop["subject"][0]
#         subphase = "Stroop{}".format(str(df_stroop["sessionid"][0])[-1])
#         df_mean = df_stroop.mean(axis=0).to_frame().T
#
#         if cols:
#             dict_stroop[subphase] = df_mean[cols]
#         else:
#             dict_stroop[subphase] = df_mean
#
#     return dict_stroop


def load_pandas_dict_excel(
    file_path: path_t, index_col: Optional[str] = "time", timezone: Optional[Union[str, datetime.tzinfo]] = None
) -> Dict[str, pd.DataFrame]:
    """Load Excel file containing pandas dataframes with time series data of one subject.

    Parameters
    ----------
    file_path : :class:`~pathlib.Path` or str
        path to file
    index_col : str, optional
        name of index columns of dataframe or ``None`` if no index column is present. Default: "time"
    timezone : str or :class:`datetime.tzinfo`, optional
        timezone of the acquired data for localization (since Excel does not support localized timestamps),
        either as string of as tzinfo object.
        Default: "Europe/Berlin"

    Returns
    -------
    dict
        dictionary with multiple pandas dataframes

    Raises
    ------
    :class:`~biopsykit.utils.exceptions.FileExtensionError`
        if file is no Excel file (".xls" or ".xlsx")

    See Also
    --------
    write_pandas_dict_excel : Write dictionary with dataframes to file

    """
    # ensure pathlib
    file_path = Path(file_path)
    _assert_file_extension(file_path, (".xls", ".xlsx"))

    # assure that the file is an Excel file
    is_excel_file(file_path)

    dict_df: Dict[str, pd.DataFrame] = pd.read_excel(file_path, index_col=index_col, sheet_name=None)

    # (re-)localize each sheet since Excel does not support timezone-aware dates (if index is DatetimeIndex)
    for key in dict_df:
        if isinstance(dict_df[key].index, pd.DatetimeIndex):
            dict_df[key] = dict_df[key].tz_localize(timezone)
    return dict_df


def write_pandas_dict_excel(
    data_dict: Dict[str, pd.DataFrame],
    file_path: path_t,
    index_col: Optional[bool] = True,
):
    """Write a dictionary with pandas dataframes to an Excel file.

    Parameters
    ----------
    data_dict : dict
        dictionary with pandas dataframes
    file_path : :class:`~pathlib.Path` or str
        path to exported Excel file
    index_col : bool, optional
        ``True`` to include dataframe index in Excel export, ``False`` otherwise. Default: ``True``

    Raises
    ------
    :exc:`~biopsykit.utils.exceptions.FileExtensionError`
        if ``file_path`` is not an Excel file

    """
    # ensure pathlib
    file_path = Path(file_path)
    _assert_file_extension(file_path, [".xls", ".xlsx"])

    writer = pd.ExcelWriter(file_path, engine="xlsxwriter")  # pylint:disable=abstract-class-instantiated
    for key in data_dict:
        if isinstance(data_dict[key].index, pd.DatetimeIndex):
            # un-localize DateTimeIndex because Excel doesn't support timezone-aware dates
            data_dict[key].tz_localize(None).to_excel(writer, sheet_name=key, index=index_col)
        else:
            data_dict[key].to_excel(writer, sheet_name=key, index=index_col)
    writer.close()


def write_result_dict(
    result_dict: Dict[str, pd.DataFrame],
    file_path: path_t,
    index_name: Optional[str] = "subject",
):
    """Write dictionary with processing results (e.g. HR, HRV, RSA) to csv file.

    The keys in the dictionary should be the subject IDs (or any other identifier),
    the values should be :class:`~pandas.DataFrame`. The index level(s) of the exported dataframe can be specified
    by the ``index_col`` parameter.

    The dictionary will be concatenated to one large dataframe which will then be saved as csv file.

    Parameters
    ----------
    result_dict : dict
        Dictionary containing processing results for all subjects. The keys in the dictionary should be the Subject IDs
        (or any other identifier), the values should be pandas dataframes
    file_path : :class:`~pathlib.Path`, str
        path to file
    index_name : str, optional
        name of the index resulting from concatenting dataframes. Default: ``subject``

    Raises
    ------
    :exc:`~biopsykit.utils.exceptions.FileExtensionError`
        if ``file_path`` is not a csv or Excel file

    Examples
    --------
    >>>
    >>> from biopsykit.io import write_result_dict
    >>>
    >>> file_path = "./param_results.csv"
    >>>
    >>> dict_param_output = {
    >>> 'S01' : pd.DataFrame(), # e.g., dataframe from mist_param_subphases,
    >>> 'S02' : pd.DataFrame(),
    >>> # ...
    >>> }
    >>>
    >>> write_result_dict(dict_param_output, file_path=file_path, index_name="subject")

    """
    # ensure pathlib
    file_path = Path(file_path)
    _assert_file_extension(file_path, [".csv", ".xls", ".xlsx"])
    df_result_concat = pd.concat(result_dict, names=[index_name])
    if file_path.suffix in [".csv"]:
        df_result_concat.to_csv(file_path)
    else:
        writer = pd.ExcelWriter(file_path, engine="xlsxwriter")  # pylint:disable=abstract-class-instantiated
        df_result_concat.to_excel(writer)
        writer.close()


def _extract_date(dataset: Dataset, df: pd.DataFrame, date: Union[str, datetime.datetime]) -> datetime.datetime:
    if dataset is not None:
        date = dataset.info.utc_datetime_start.date()
    if df is not None:
        if isinstance(df.index, pd.DatetimeIndex):
            date = df.index.normalize().unique()[0]
            date = date.to_pydatetime()
        else:
            raise ValueError("'df' must have a DatetimeIndex!")
    if isinstance(date, str):
        # ensure datetime
        date = pd.to_datetime(date)
        date = date.date()

    return date


def _get_subject_col(data: pd.DataFrame, subject_col: str):
    if subject_col is None:
        subject_col = "subject"
    _assert_is_dtype(subject_col, str)
    _assert_has_columns(data, [[subject_col]])
    return subject_col


def _sanitize_index_cols(
    data: pd.DataFrame,
    subject_col: str,
    condition_col: Optional[str],
    additional_index_cols: Optional[Union[str, Sequence[str]]],
) -> Tuple[pd.DataFrame, Sequence[str]]:
    subject_col = _get_subject_col(data, subject_col)
    data = data.rename(columns={subject_col: "subject"})
    subject_col = "subject"
    index_cols = [subject_col]

    if condition_col is not None:
        _assert_is_dtype(condition_col, str)
        _assert_has_columns(data, [[condition_col]])
        data = data.rename(columns={condition_col: "condition"})
        condition_col = "condition"
        index_cols.append(condition_col)
    elif "condition" in data.columns:
        index_cols.append("condition")

    if additional_index_cols is None:
        additional_index_cols = []
    if isinstance(additional_index_cols, str):
        additional_index_cols = [additional_index_cols]

    index_cols = index_cols + additional_index_cols
    return data, index_cols


def _load_dataframe(file_path, **kwargs):
    if file_path.suffix in [".csv"]:
        return pd.read_csv(file_path, **kwargs)
    return pd.read_excel(file_path, **kwargs)


def _apply_index_cols(
    data: pd.DataFrame, index_cols: Optional[Union[str, Sequence[str], Dict[str, str]]] = None
) -> pd.DataFrame:
    new_index_cols = None
    if isinstance(index_cols, str):
        index_cols = [index_cols]
    elif isinstance(index_cols, dict):
        new_index_cols = list(index_cols.values())
        index_cols = list(index_cols.keys())

    if index_cols is not None:
        _assert_has_columns(data, [index_cols])
        data = data.set_index(index_cols)

    if new_index_cols is not None:
        data.index = data.index.set_names(new_index_cols)

    return data


def _apply_phase_cols(data: pd.DataFrame, phase_cols: Union[Dict[str, Sequence[str]], Sequence[str]]) -> pd.DataFrame:
    new_phase_cols = None
    if isinstance(phase_cols, dict):
        new_phase_cols = phase_cols
        phase_cols = list(phase_cols.keys())

    if phase_cols:
        _assert_has_columns(data, [phase_cols])
        data = data.loc[:, phase_cols]
    if new_phase_cols:
        data = data.rename(columns=new_phase_cols)

    return data


def _parse_time_log_not_continuous(
    data: pd.DataFrame, index_cols: Union[str, Sequence[str], Dict[str, str]]
) -> pd.DataFrame:
    start_cols = np.squeeze(data.columns.str.extract(r"(\w+)_start").dropna().values)
    end_cols = np.squeeze(data.columns.str.extract(r"(\w+)_end").dropna().values)
    if start_cols.size == 0:
        raise ValidationError(
            "No 'start' and 'end' columns were found. "
            "Make sure that each phase has columns with 'start' and 'end' suffixes!"
        )
    if not np.array_equal(start_cols, end_cols):
        raise ValidationError("Not all phases have 'start' and 'end' columns!")

    if index_cols is None:
        index_cols = [s for s in ["subject", "condition"] if s in data.columns]
        data = data.set_index(index_cols)
    if isinstance(index_cols, dict):
        index_cols = data.index.names

    data = pd.wide_to_long(
        data.reset_index(),
        stubnames=start_cols,
        i=index_cols,
        j="time",
        sep="_",
        suffix="(start|end)",
    )

    # ensure that "start" is always before "end"
    data = data.reindex(["start", "end"], level=-1)
    # unstack start|end level
    data = data.unstack()
    # set name of outer index level
    data.columns = data.columns.set_names("phase", level=0)

    return data



"""Module providing some standard plots for visualizing data collected during a psychological protocol."""
import re
from typing import Any, Dict, Iterable, Optional, Sequence, Tuple, Union

import matplotlib.patches as mpatch
import matplotlib.pyplot as plt
import matplotlib.ticker as mticks
import numpy as np
import pandas as pd
import seaborn as sns
from biopsykit.plotting import feature_boxplot, lineplot, multi_feature_boxplot
from biopsykit.protocols._utils import _get_sample_times
from biopsykit.saliva.utils import _remove_s0
from biopsykit.utils.data_processing import get_subphase_durations
from biopsykit.utils.datatype_helper import (
    MeanSeDataFrame,
    MergedStudyDataDict,
    SalivaFeatureDataFrame,
    SalivaMeanSeDataFrame,
    SalivaRawDataFrame,
    is_mean_se_dataframe,
    is_saliva_feature_dataframe,
    is_saliva_mean_se_dataframe,
    is_saliva_raw_dataframe,
)
from biopsykit.utils.exceptions import ValidationError
from fau_colors import cmaps, colors_all
from matplotlib.legend_handler import HandlerTuple

_hr_ensemble_plot_params = {
    "linestyle": ["solid", "dashed", "dotted", "dashdot"],
    "ensemble_alpha": 0.4,
    "background_base_color": "#e0e0e0",
    "background_color": None,
    "background_alpha": 0.1,
    "xlabel": r"Time [s]",
    "xaxis_minor_tick_locator": mticks.MultipleLocator(60),
    "ylabel": "Heart Rate [bpm]",
    "legend_loc": "lower right",
    "legend_bbox_to_anchor": (0.99, 0.01),
    "phase_text": "{}",
    "end_phase_text": "End {}",
    "end_phase_line_color": "#e0e0e0",
    "end_phase_line_style": "--",
    "end_phase_line_width": 2.0,
}

_hr_mean_plot_params = {
    "linestyle": ["solid", "dashed", "dotted", "dashdot"],
    "marker": ["o", "P", "*", "X"],
    "background_base_color": "#e0e0e0",
    "background_color": None,
    "background_alpha": 0.1,
    "x_offset": 0.1,
    "ylabel": r"Heart Rate [bpm]",
    "phase_text": "{}",
}

_saliva_feature_params: Dict[str, Dict[str, Any]] = {
    "ylabel": {
        "cortisol": {
            "auc": r"Cortisol AUC $\left[\frac{nmol \cdot min}{l} \right]$",
            "auc_g": r"Cortisol AUC $\left[\frac{nmol \cdot min}{l} \right]$",
            "auc_i": r"Cortisol AUC $\left[\frac{nmol \cdot min}{l} \right]$",
            "auc_i_post": r"Cortisol AUC $\left[\frac{nmol \cdot min}{l} \right]$",
            "slope": r"Cortisol Change $\left[\frac{nmol}{l \cdot min} \right]$",
            "max": r"Cortisol $\left[\frac{nmol}{l} \right]$",
            "argmax": r"Cortisol $\left[\frac{nmol}{l} \right]$",
            "max_inc": r"Cortisol $\left[\frac{nmol}{l} \right]$",
            "mean": r"Cortisol $\left[\frac{nmol}{l} \right]$",
            "std": r"Cortisol $\left[\frac{nmol}{l} \right]$",
            "kurt": r"Cortisol $\left[\frac{nmol}{l} \right]$",
            "skew": r"Cortisol $\left[\frac{nmol}{l} \right]$",
        },
        "amylase": {
            "auc": r"Amylase AUC $\left[\frac{U \cdot min}{l} \right]$",
            "auc_g": r"Amylase AUC $\left[\frac{U \cdot min}{l} \right]$",
            "auc_i": r"Amylase AUC $\left[\frac{U \cdot min}{l} \right]$",
            "auc_i_post": r"Amylase AUC $\left[\frac{U \cdot min}{l} \right]$",
            "slope": r"Amylase Change $\left[\frac{U}{l \cdot min} \right]$",
            "max": r"Amylase $\left[\frac{U}{l} \right]$",
            "max_inc": r"Amylase $\left[\frac{U}{l} \right]$",
            "mean": r"Amylase $\left[\frac{U}{l} \right]$",
            "std": r"Amylase $\left[\frac{U}{l} \right]$",
            "kurt": r"Amylase $\left[\frac{U}{l} \right]$",
            "skew": r"Amylase $\left[\frac{U}{l} \right]$",
        },
    },
    "xticklabels": {
        "auc": r"$AUC_{$}$",
        "auc_g": r"$AUC_G$",
        "auc_i": r"$AUC_I$",
        "auc_i_post": r"$AUC_{I}^{Post}$",
        "slope": r"$a_{§}$",
        "max_inc": r"$\Delta c_{max}$",
        "cmax": r"$c_{max}$",
        "argmax": r"$argmax(c)$",
        "mean": r"$\mu(c)$",
        "std": r"$\sigma(c)$",
        "skew": r"$skew(c)$",
        "kurt": r"$kurt(c)$",
    },
}

_saliva_plot_params: Dict = {
    "palette": None,
    "linestyle": ["-", "--"],
    "marker": ["o", "P"],
    "test_title": "",
    "test_fontsize": "medium",
    "test_color": "#9e9e9e",
    "test_alpha": 0.2,
    "multi_x_offset": 0.01,
    "xlabel": "Time [min]",
    "ylabel": {
        "cortisol": "Cortisol [nmol/l]",
        "amylase": "sAA [U/l]",
        "il6": "IL-6 [pg/ml]",
    },
    "legend_title": {"cortisol": "Cortisol", "amylase": "sAA", "il6": "IL-6"},
}


def _get_palette(color: Optional[Union[str, Sequence[str]]] = None, num_colors: Optional[int] = 3):
    if isinstance(color, list):
        return color
    if color is None:
        color = "fau"
    color_val = getattr(colors_all, color, None)
    if color_val is None:
        return color
    return sns.light_palette(color_val, num_colors + 1, reverse=True)[:-1]


def hr_ensemble_plot(
    data: MergedStudyDataDict,
    subphases: Optional[Dict[str, Dict[str, int]]] = None,
    **kwargs,
) -> Optional[Tuple[plt.Figure, plt.Axes]]:
    r"""Draw a heart rate ensemble plot.

    This function plots time-series heart rate continuously as ensemble plot (mean ± standard error).
    If the data consist of multiple phases, data from each phase are overlaid in the same plot.
    If each phase additionally consists of subphases, the single subphases are highlighted in the plot.

    The input data is expected to be a :obj:`~biopsykit.utils.datatype_helper.MergedStudyDataDict`, i.e.,
    a dictionary with merged time-series heart rate data, of multiple subjects, split into individual phases.
    Per phase, the data of each subjects have same length and are combined into one common dataframe.


    Parameters
    ----------
    data : :obj:`~biopsykit.utils.datatype_helper.MergedStudyDataDict`
        dict with heart rate data to plot
    subphases : dict, optional
        dictionary with phases (keys) and subphases (values - dict with subphase names and subphase durations) or
        ``None`` if no subphases are present. Default: ``None``
    **kwargs : dict, optional
        optional arguments for plot configuration.

        To style general plot appearance:

        * ``ax``: pre-existing axes for the plot. Otherwise, a new figure and axes object is created and returned.
        * ``palette``: color palette to plot data from different phases
        * ``ensemble_alpha``: transparency value for ensemble plot errorband (around mean). Default: 0.3
        * ``background_alpha``: transparency value for background spans (if subphases are present). Default: 0.2
        * ``linestyle``: list of line styles for ensemble plots. Must match the number of phases to plot
        * ``phase_text``: string pattern to customize phase name shown in legend with placeholder for subphase name.
          Default: "{}"

        To style axes:

        * ``is_relative``: boolean indicating whether heart rate data is relative (in % relative to baseline)
          or absolute (in bpm). Default: ``True``
        * ``xlabel``: label of x axis. Default: ":math:`Time [s]`"
        * ``xaxis_minor_tick_locator``: locator object to style x axis minor ticks. Default: 60 sec
        * ``ylabel``: label of y axis. Default: ":math:`\Delta HR [\%]`"
        * ``ylims``: y axis limits. Default: ``None`` to automatically infer limits

        To style the annotations at the end of each phase:

        * ``end_phase_text``: string pattern to customize text at the end of phase with placeholder for phase name.
          Default: "{}"
        * ``end_phase_line_color``: line color of vertical lines used to indicate end of phase. Default: "#e0e0e0"
        * ``end_phase_line_width``: line width of vertical lines used  to indicate end of phase. Default: 2.0

        To style legend:

        * ``legend_loc``: location of legend. Default: "lower right"
        * ``legend_bbox_to_anchor``: box that is used to position the legend in conjunction with ``legend_loc``


    Returns
    -------
    fig : :class:`~matplotlib.figure.Figure`
        figure object
    ax : :class:`~matplotlib.axes.Axes`
        axes object


    See Also
    --------
    :obj:`~biopsykit.utils.datatype_helper.MergedStudyDataDict`
        dictionary format
    :func:`~biopsykit.utils.data_processing.merge_study_data_dict`
        function to build ``MergedStudyDataDict``


    Examples
    --------
    >>> from biopsykit.protocols.plotting import hr_ensemble_plot
    >>> # Example with subphases
    >>> subphase_dict = {
    >>>     "Phase1": {"Baseline": 60, "Stress": 120, "Recovery": 60},
    >>>     "Phase2": {"Baseline": 60, "Stress": 120, "Recovery": 60},
    >>>     "Phase3": {"Baseline": 60, "Stress": 120, "Recovery": 60}
    >>> }
    >>> fig, ax = hr_ensemble_plot(data=data, subphases=subphase_dict)

    """
    ax: plt.Axes = kwargs.pop("ax", None)
    if ax is None:
        fig, ax = plt.subplots(figsize=kwargs.get("figsize"))
    else:
        fig = ax.get_figure()

    palette = kwargs.get("palette")
    palette = _get_palette(palette, len(data))
    sns.set_palette(palette)

    linestyle = kwargs.get("linestyle", _hr_ensemble_plot_params.get("linestyle"))

    xlabel = kwargs.get("xlabel", _hr_ensemble_plot_params.get("xlabel"))
    ylabel_default = _hr_ensemble_plot_params.get("ylabel")
    if kwargs.get("is_relative", True):
        ylabel_default = r"$\Delta$ HR [%]"
    ylabel = kwargs.get("ylabel", ylabel_default)
    ylims = kwargs.get("ylims", _hr_ensemble_plot_params.get("ylims"))
    xaxis_minor_tick_locator = kwargs.get(
        "xaxis_minor_tick_locator", _hr_ensemble_plot_params.get("xaxis_minor_tick_locator")
    )

    ensemble_alpha = kwargs.get("ensemble_alpha", _hr_ensemble_plot_params.get("ensemble_alpha"))
    phase_text = kwargs.get("phase_text", _hr_ensemble_plot_params.get("phase_text"))

    legend_loc = kwargs.get("legend_loc", _hr_ensemble_plot_params.get("legend_loc"))
    legend_bbox_to_anchor = kwargs.get("legend_bbox_to_anchor", _hr_ensemble_plot_params.get("legend_bbox_to_anchor"))

    for i, phase in enumerate(data):
        df_hr_phase = data[phase]
        x = df_hr_phase.index
        hr_mean = df_hr_phase.mean(axis=1)
        hr_stderr = df_hr_phase.std(axis=1) / np.sqrt(df_hr_phase.shape[1])
        ax.plot(x, hr_mean, zorder=2, label=phase_text.format(phase), linestyle=linestyle[i])
        ax.fill_between(x, hr_mean - hr_stderr, hr_mean + hr_stderr, zorder=1, alpha=ensemble_alpha)
        _hr_ensemble_plot_end_phase_annotation(ax, df_hr_phase, phase, i, **kwargs)

    if subphases is not None:
        _hr_ensemble_plot_subphase_vspans(ax, data, subphases, **kwargs)

    ax.set_xlabel(xlabel)
    ax.xaxis.set_minor_locator(xaxis_minor_tick_locator)
    ax.tick_params(axis="x", which="both", bottom=True)

    ax.set_ylabel(ylabel)
    ax.tick_params(axis="y", which="major", left=True)

    if ylims is not None:
        ax.margins(x=0)
        ax.set_ylim(ylims)
    else:
        ax.margins(0, 0.1)

    ax.legend(loc=legend_loc, bbox_to_anchor=legend_bbox_to_anchor)

    fig.tight_layout()
    return fig, ax


def _hr_ensemble_plot_end_phase_annotation(ax: plt.Axes, data: pd.DataFrame, phase: str, i: int, **kwargs):
    """Add End Phase annotations to heart rate ensemble plot.

    Parameters
    ----------
    ax : :class:`matplotlib.axes.Axes`
        axes object
    data : :class:`~pandas.DataFrame`
        data belonging to ``phase``
    phase : str
        phase to add annotations
    i : int
        counter of phase

    """
    end_phase_text = kwargs.get("end_phase_text", _hr_ensemble_plot_params.get("end_phase_text"))
    end_phase_line_color = kwargs.get("end_phase_line_color", _hr_ensemble_plot_params.get("end_phase_line_color"))
    end_phase_line_style = kwargs.get("end_phase_line_style", _hr_ensemble_plot_params.get("end_phase_line_style"))
    end_phase_line_width = kwargs.get("end_phase_line_width", _hr_ensemble_plot_params.get("end_phase_line_width"))

    ax.vlines(
        x=len(data),
        ymin=0,
        ymax=1,
        transform=ax.get_xaxis_transform(),
        ls=end_phase_line_style,
        lw=end_phase_line_width,
        colors=end_phase_line_color,
        zorder=3,
    )
    ax.annotate(
        text=end_phase_text.format(phase),
        xy=(len(data), 0.85 - 0.075 * i),
        xytext=(-5, 0),
        xycoords=ax.get_xaxis_transform(),
        textcoords="offset points",
        ha="right",
        fontsize="small",
        bbox={"facecolor": "#e0e0e0", "alpha": 0.7, "boxstyle": "round"},
        zorder=5,
    )


def _hr_ensemble_plot_subphase_vspans(
    ax: plt.Axes, data: Dict[str, pd.DataFrame], subphases: Dict[str, Dict[str, int]], **kwargs
):
    """Add subphase vertical spans (vspans) to heart rate ensemble plot.

    Parameters
    ----------
    ax : :class:`matplotlib.axes.Axes`
        axes object
    data : :class:`~pandas.DataFrame`
        data belonging to ``phase``
    subphases : dict
        dictionary with phases (keys) and subphases (values - dict with subphase names and subphase durations)

    """
    subphase_times = [get_subphase_durations(df, subphases[phase]) for phase, df in data.items()]
    subphase_times = np.array(subphase_times)
    subphase_times = np.max(subphase_times, axis=0)

    subphase_names = np.array([list(subphase_dict.keys()) for phase, subphase_dict in subphases.items()])

    if not (subphase_names[0] == subphase_names).all():
        raise ValueError("Subphases must be the same for all phases!")
    subphase_names = subphase_names[0]

    bg_colors = kwargs.get("background_color", _hr_ensemble_plot_params.get("background_color"))
    if bg_colors is None:
        bg_color_base = kwargs.get("background_base_color", _hr_ensemble_plot_params.get("background_base_color"))
        bg_colors = list(sns.dark_palette(bg_color_base, n_colors=len(subphase_names), reverse=True))
    bg_alphas = kwargs.get("background_alpha", _hr_ensemble_plot_params.get("background_alpha"))
    bg_alphas = [bg_alphas] * len(subphase_names)

    for i, subphase in enumerate(subphase_names):
        start, end = subphase_times[i]
        color = bg_colors[i]
        alpha = bg_alphas[i]
        ax.axvspan(start, end, color=color, alpha=alpha, zorder=0, lw=0)
        ax.text(
            x=start + 0.5 * (end - start),
            y=0.95,
            transform=ax.get_xaxis_transform(),
            zorder=3,
            s=subphase,
            ha="center",
            va="center",
        )
    p = mpatch.Rectangle(
        xy=(0, 0.9),
        width=1,
        height=0.1,
        transform=ax.transAxes,
        color="white",
        alpha=0.4,
        zorder=1,
        lw=0,
    )
    ax.add_patch(p)
    ax.set_xticks([start for (start, end) in subphase_times])


def hr_mean_plot(  # pylint:disable=too-many-branches
    data: MeanSeDataFrame,
    **kwargs,
) -> Tuple[plt.Figure, plt.Axes]:
    r"""Plot course of heart rate as mean ± standard error over phases (and subphases) of a psychological protocol.

    The correct plot is automatically inferred from the provided data:

    * only ``phase`` index level: plot phases over x axis
    * ``phase`` and ``subphase`` index levels: plot subphases over x axis, highlight phases as vertical spans
    * additionally: ``condition`` level: plot data of different conditions individually
      (corresponds to ``hue`` parameter in :func:`~biopsykit.plotting.lineplot`)

    Parameters
    ----------
    data : :class:`~biopsykit.utils.datatype_helper.MeanSeDataFrame`
        Heart rate data to plot. Must be provided as ``MeanSeDataFrame`` with columns ``mean`` and ``se``
        computed over phases (and, if available, subphases)
    **kwargs
        additional parameters to be passed to the plot, such as:

        * ``ax``: pre-existing axes for the plot. Otherwise, a new figure and axes object is created and returned.
        * ``figsize``: tuple specifying figure dimensions
        * ``palette``: color palette to plot data from different conditions. If ``palette`` is a str then it is
          assumed to be the name of a ``fau_colors`` palette (``fau_colors.cmaps._fields``).
        * ``is_relative``: boolean indicating whether heart rate data is relative (in % relative to baseline)
          or absolute (in bpm). Default: ``False``
        * ``order``: list specifying the order of categorical values (i.e., conditions) along the x axis.
        * ``x_offset``: offset value to move different groups along the x axis for better visualization.
          Default: 0.05
        * ``xlabel``: label of x axis. Default: "Subphases" (if subphases are present)
          or "Phases" (if only phases are present).
        * ``ylabel``: label of y axis. Default: ":math:`\Delta HR [%]`"
        * ``ylims``: list to manually specify y axis limits, float to specify y axis margin
          (see :meth:`~matplotlib.axes.Axes.margins()` for further information), or ``None`` to automatically infer
          y axis limits.
        * ``marker``: string or list of strings to specify marker style.
          If ``marker`` is a string, then marker of each line will have the same style.
          If ``marker`` is a list, then marker of each line will have a different style.
        * ``linestyle``: string or list of strings to specify line style.
          If ``linestyle`` is a string, then each line will have the same style.
          If ``linestyle`` is a list, then each line will have a different style.

    Returns
    -------
    fig : :class:`~matplotlib.figure.Figure`
        figure object
    ax : :class:`~matplotlib.axes.Axes`
        axes object

    See Also
    --------
    :func:`~biopsykit.plotting.lineplot`
        Plot data as lineplot with mean and standard error

    """
    fig, ax = _plot_get_fig_ax(**kwargs)
    kwargs.update({"ax": ax})

    num_conditions = 1
    if "condition" in data.index.names:
        num_conditions = len(data.index.names)

    # get all plot parameter
    palette = kwargs.get("palette", cmaps.faculties)
    palette = _get_palette(palette, num_conditions)
    sns.set_palette(palette)

    ylabel_default = _hr_mean_plot_params.get("ylabel")
    if kwargs.get("is_relative", False):
        ylabel_default = r"$\Delta$ HR [%]"
    ylabel = kwargs.get("ylabel", ylabel_default)
    ylims = kwargs.get("ylims", None)

    phase_dict = _hr_mean_get_phases_subphases(data)
    num_phases = len(phase_dict)
    num_subphases = [len(arr) for arr in phase_dict.values()]
    x_vals = _hr_mean_get_x_vals(num_phases, num_subphases)

    # build x axis, axis limits and limits for phase spans
    dist = np.mean(np.ediff1d(x_vals))
    x_lims = np.append(x_vals, x_vals[-1] + dist)
    x_lims = x_lims - 0.5 * np.ediff1d(x_lims, to_end=dist)

    if "condition" in data.index.names:
        data_grp = dict(tuple(data.groupby("condition")))
        order = kwargs.get("order", list(data_grp.keys()))
        data_grp = {key: data_grp[key] for key in order}

        for i, (key, df) in enumerate(data_grp.items()):
            _hr_mean_plot(df, x_vals, key, index=i, **kwargs)
    else:
        _hr_mean_plot(data, x_vals, "Data", index=0, **kwargs)

    # add decorators to phases if subphases are present
    if sum(num_subphases) > 0:
        _hr_mean_plot_subphase_annotations(phase_dict, x_lims, **kwargs)

    # customize x axis
    ax.tick_params(axis="x", bottom=True)
    ax.set_xticks(x_vals)
    ax.set_xlim(np.min(x_lims), np.max(x_lims))
    _hr_mean_style_x_axis(ax, phase_dict, num_subphases)

    # customize y axis
    ax.tick_params(axis="y", which="major", left=True)
    ax.set_ylabel(ylabel)

    _hr_mean_plot_set_axis_lims(ylims, ax)

    # customize legend
    if "condition" in data.index.names:
        _hr_mean_add_legend(**kwargs)

    fig.tight_layout()
    return fig, ax


def _hr_mean_plot_set_axis_lims(ylims: Union[Sequence[float], float], ax: plt.Axes):
    if isinstance(ylims, (tuple, list)):
        ax.set_ylim(ylims)
    else:
        ymargin = 0.15
        if isinstance(ylims, float):
            ymargin = ylims
        ax.margins(y=ymargin)
    ax.margins(x=0)
    ax.relim()


def _hr_mean_plot(data: MeanSeDataFrame, x_vals: np.array, key: str, index: int, **kwargs):
    ax: plt.Axes = kwargs.get("ax")
    x_offset = kwargs.get("x_offset", _hr_mean_plot_params.get("x_offset"))
    marker = kwargs.get("marker", _hr_mean_plot_params.get("marker"))
    linestyle = kwargs.get("linestyle", _hr_mean_plot_params.get("linestyle"))

    if isinstance(marker, list):
        marker = marker[index]

    if isinstance(linestyle, list):
        linestyle = linestyle[index]

    is_mean_se_dataframe(data)
    if isinstance(data.columns, pd.MultiIndex):
        # if data has multiindex columns: drop all levels except the last one
        # (which is expected to contain ["mean", "se"])
        data.columns = data.columns.droplevel(list(range(0, data.columns.nlevels - 1)))

    ax.errorbar(
        x=x_vals + index * x_offset,
        y=data["mean"],
        label=key,
        yerr=data["se"],
        capsize=3,
        marker=marker,
        linestyle=linestyle,
    )


def _hr_mean_add_legend(**kwargs):
    """Add legend to mean HR plot."""
    ax: plt.Axes = kwargs.get("ax")
    legend_loc = kwargs.get("legend_loc", "upper left")
    # get handles
    handles, labels = ax.get_legend_handles_labels()
    # remove the errorbars
    handles = [h[0] for h in handles]
    # use them in the legend
    if legend_loc == "upper left":
        bbox_to_anchor = (0.01, 0.90)
    elif legend_loc == "upper right":
        bbox_to_anchor = (0.99, 0.90)
    else:
        bbox_to_anchor = None
    ax.legend(
        handles,
        labels,
        loc=legend_loc,
        bbox_to_anchor=bbox_to_anchor,
        numpoints=1,
    )


def _hr_mean_style_x_axis(ax: plt.Axes, phase_dict: Dict[str, Sequence[str]], num_subphases: Sequence[int], **kwargs):
    """Style x axis of mean HR plot.

    Parameters
    ----------
    ax : :class:`~matplotlib.axes.Axes`
        axes object
    phase_dict : dict
        dictionary with phase names (keys) and dict of subphases (values)
    num_subphases : list
        list with number of subphases for each phase

    """
    if sum(num_subphases) == 0:
        # no subphases
        ax.set_xticklabels(phase_dict.keys())
        ax.set_xlabel(kwargs.get("xlabel", "Phases"))
    else:
        ax.set_xticklabels([s for subph in phase_dict.values() for s in subph])
        ax.set_xlabel(kwargs.get("xlabel", "Subphases"))


def _hr_mean_plot_subphase_annotations(phase_dict: Dict[str, Sequence[str]], xlims: Sequence[float], **kwargs):
    """Add subphase annotations to mean HR plot.

    Parameters
    ----------
    phase_dict : dict
        dictionary with phase names (keys) and dict of subphases (values)
    xlims : list
        x axis limits

    """
    ax: plt.Axes = kwargs.get("ax")

    num_phases = len(phase_dict)
    num_subphases = [len(arr) for arr in phase_dict.values()]

    bg_colors = kwargs.get("background_color", _hr_ensemble_plot_params.get("background_color"))
    if bg_colors is None:
        bg_color_base = kwargs.get("background_base_color", _hr_ensemble_plot_params.get("background_base_color"))
        bg_colors = list(sns.dark_palette(bg_color_base, n_colors=num_phases, reverse=True))
    bg_alphas = kwargs.get("background_alpha", _hr_ensemble_plot_params.get("background_alpha"))
    bg_alphas = [bg_alphas] * num_phases

    phase_text = kwargs.get("phase_text", _hr_mean_plot_params.get("phase_text"))

    x_spans = _hr_mean_get_x_spans(num_phases, num_subphases)

    for (i, phase) in enumerate(phase_dict):
        left, right = x_spans[i]
        bg_color = bg_colors[i]
        bg_alpha = bg_alphas[i]
        ax.axvspan(xlims[left], xlims[right], color=bg_color, alpha=bg_alpha, zorder=0, lw=0)
        name = phase_text.format(phase)
        ax.text(
            x=xlims[left] + 0.5 * (xlims[right] - xlims[left]),
            y=0.95,
            s=name,
            transform=ax.get_xaxis_transform(),
            horizontalalignment="center",
            verticalalignment="center",
            zorder=3,
        )

    p = mpatch.Rectangle(
        xy=(0, 0.9),
        width=1,
        height=0.1,
        transform=ax.transAxes,
        color="white",
        alpha=0.4,
        zorder=1,
        lw=0,
    )
    ax.add_patch(p)


def _hr_mean_get_x_spans(num_phases: int, num_subphases: Sequence[int]):
    if sum(num_subphases) == 0:
        x_spans = list(zip([0, *list(range(0, num_phases))], list(range(0, num_phases))))
    else:
        x_spans = list(zip([0, *list(np.cumsum(num_subphases))], list(np.cumsum(num_subphases))))
    return x_spans


def _hr_mean_get_x_vals(num_phases: int, num_subphases: Sequence[int]):
    x_vals = np.linspace(0, 10, num_phases) if sum(num_subphases) == 0 else np.linspace(0, 10, sum(num_subphases))
    return x_vals


def _hr_mean_get_phases_subphases(data: pd.DataFrame) -> Dict[str, Sequence[str]]:
    if "condition" in data.index.names:
        data = [value for key, value in data.groupby("condition")][0]

    phases = data.index.get_level_values("phase").unique()

    if "subphase" in data.index.names:
        phase_dict = {phase: list(df.index.get_level_values("subphase")) for phase, df in data.groupby("phase")}
    else:
        phase_dict = {phase: [] for phase in phases}

    return phase_dict


def saliva_plot(  # pylint:disable=too-many-branches
    data: Union[
        SalivaRawDataFrame, SalivaMeanSeDataFrame, Dict[str, SalivaRawDataFrame], Dict[str, SalivaMeanSeDataFrame]
    ],
    saliva_type: Optional[str] = None,
    sample_times: Optional[Union[Sequence[int], Dict[str, Sequence[int]]]] = None,
    test_times: Optional[Sequence[int]] = None,
    sample_times_absolute: Optional[bool] = False,
    remove_s0: Optional[bool] = False,
    **kwargs,
) -> Optional[Tuple[plt.Figure, plt.Axes]]:
    r"""Plot saliva data during psychological protocol as mean ± standard error.

    The function accepts raw saliva data per subject (:obj:`~biopsykit.utils.datatype_helper.SalivaRawDataFrame`)
    as well as pre-computed mean and standard error values of saliva samples
    ( :obj:`~biopsykit.utils.datatype_helper.SalivaMeanSeDataFrame`). To combine data from multiple saliva
    types (maximum: 2) into one plot a dict can be passed to ``data``.

    If a psychological test (e.g., TSST, MIST, or Stroop) was performed, the test time is highlighted as vertical span
    within the plot.

    .. note::
        If no sample times are provided (neither via ``time`` column in ``data`` nor via ``sample_times`` parameter)
        then ``samples`` will be used as x axis


    Parameters
    ----------
    data : :obj:`~biopsykit.utils.datatype_helper.SalivaRawDataFrame`, \
           :obj:`~biopsykit.utils.datatype_helper.SalivaMeanSeDataFrame`, or dict of such
        Saliva data to plot. Must either be provided as ``SalivaRawDataFrame`` with raw saliva data per subject or
        as ``SalivaMeanSeDataFrame`` with columns ``mean`` and ``se`` computed per saliva sample. To plot data from
        multiple saliva types (maximum: 2) a dict can be passed (keys: saliva types, values: saliva data).
    saliva_type : {"cortisol", "amylase", "il6"}, optional
        saliva type to be plotted. If a dict is passed and ``saliva_type`` is ``None``
        the saliva types are inferred from dict keys.
    sample_times : list or dict of lists
        sample times in minutes relative to psychological test or a dict of such if sample times are different for
        the individual saliva types.
    test_times : list of int, optional
        start and end times of psychological test (in minutes) or ``None`` if no test was performed
    sample_times_absolute : bool, optional
        ``True`` if absolute sample times were provided (i.e., the duration of the psychological test was already
        added to the sample times), ``False`` if relative sample times were provided and absolute times should be
        computed based on test times specified by ``test_times``. Default: ``False``
    remove_s0 : bool, optional
        whether to remove the first saliva sample for plotting or not. Default: ``False``
    **kwargs
        additional parameters to be passed to the plot.

        To style general plot appearance:

        * ``ax``: pre-existing axes for the plot. Otherwise, a new figure and axes object is created and returned.
        * ``palette``: color palette to plot data from different phases
        * ``figsize``: tuple specifying figure dimensions
        * ``marker``: string or list of strings to specify marker style.
          If ``marker`` is a string, then the markers of each line will have the same style.
          If ``marker`` is a list, then the markers of each line will have a different style.
        * ``linestyle``: string or list of strings to specify line style.
          If ``linestyle`` is a string, then each line will have the same style.
          If ``linestyle`` is a list, then each line will have a different style.

        To style axes:

        * ``x_offset``: offset value to move different groups along the x axis for better visualization.
          Default: 0.05
        * ``xlabel``: label of x axis. Default: "Subphases" (if subphases are present).
          or "Phases" (if only phases are present)
        * ``ylabel``: label of y axis. Default: ":math:`\Delta HR [%]`"
        * ``ylims``: list to manually specify y axis limits, float to specify y axis margin
          (see :meth:`~matplotlib.axes.Axes.margins()` for further information), or ``None`` to automatically infer
          y axis limits.

        To style the vertical span highlighting the psychological test in the plot:

        * ``test_title``: title of test
        * ``test_fontsize``: fontsize of the test title. Default: "medium"
        * ``test_color``: color of vspan. Default: #9e9e9e
        * ``test_alpha``: transparency value of vspan: Default: 0.5

    Returns
    -------
    fig : :class:`~matplotlib.figure.Figure`
        figure object
    ax : :class:`~matplotlib.axes.Axes`
        axes object

    See Also
    --------
    :func:`~biopsykit.plotting.lineplot`
        Plot data as lineplot with mean and standard error

    """
    fig, ax = _plot_get_fig_ax(**kwargs)
    kwargs.update({"ax": ax})

    if saliva_type is None and not isinstance(data, dict):
        raise ValueError("If 'saliva_type' is None, you must pass a dict!")

    if isinstance(data, pd.DataFrame):
        # multiple saliva data were passed in a dict => get the selected saliva type
        data = {saliva_type: data}
        sample_times = {saliva_type: sample_times}

    linestyle = kwargs.pop("linestyle", None)
    marker = kwargs.pop("marker", "o")
    palette = kwargs.pop("palette", None)
    if isinstance(palette, str) and getattr(colors_all, palette, None):
        palette = _get_palette(palette, len(data))

    for i, key in enumerate(data):
        df = data[key]
        if remove_s0:
            df = _remove_s0(df)
        if sample_times is None:
            st = None
        else:
            st = sample_times[key]
            if remove_s0:
                st = st[1:]
        kwargs_copy = _saliva_plot_extract_style_params(key, linestyle, marker, palette, **kwargs)
        _saliva_plot(
            data=df,
            saliva_type=key,
            counter=i,
            sample_times=st,
            test_times=test_times,
            sample_times_absolute=sample_times_absolute,
            **kwargs_copy,
        )

    test_times = test_times or [0, 0]
    test_title = kwargs.get("test_title", _saliva_plot_params.get("test_title"))
    test_color = kwargs.get("test_color", _saliva_plot_params.get("test_color"))
    test_alpha = kwargs.get("test_alpha", _saliva_plot_params.get("test_alpha"))
    test_fontsize = kwargs.get("test_fontsize", _saliva_plot_params.get("test_fontsize"))
    if sum(test_times) != 0:
        ax.axvspan(*test_times, color=test_color, alpha=test_alpha, zorder=1, lw=0)
        ax.text(
            x=test_times[0] + 0.5 * (test_times[1] - test_times[0]),
            y=0.95,
            transform=ax.get_xaxis_transform(),
            s=test_title,
            fontsize=test_fontsize,
            horizontalalignment="center",
            verticalalignment="top",
        )

    if len(data) > 1:
        saliva_plot_combine_legend(fig, saliva_types=list(data.keys()), **kwargs)
    else:
        fig.tight_layout()

    return fig, ax


def _saliva_plot_extract_style_params(
    key: str,
    linestyle: Union[Dict[str, str], str],
    marker: Union[Dict[str, str], str],
    palette: Union[Dict[str, str], str],
    **kwargs,
):
    ls = _saliva_plot_get_plot_param(linestyle, key)
    if linestyle is not None:
        kwargs.setdefault("linestyle", ls)

    m = _saliva_plot_get_plot_param(marker, key)
    if marker is not None:
        kwargs.setdefault("marker", m)

    cmap = _saliva_plot_get_plot_param(palette, key)
    if palette is not None:
        kwargs.setdefault("palette", cmap)

    return kwargs


def _saliva_plot_sanitize_dicts(
    data: Union[Dict[str, pd.DataFrame], pd.DataFrame], ylabel: Union[Dict[str, str], str], saliva_type: str
):
    if isinstance(ylabel, dict):
        ylabel = ylabel[saliva_type]

    if isinstance(data, dict):
        # multiple saliva data were passed in a dict => get the selected saliva type
        data = data[saliva_type]

    return data, ylabel


def _saliva_plot(
    data: Union[SalivaRawDataFrame, SalivaMeanSeDataFrame],
    saliva_type: str,
    counter: int,
    sample_times: Optional[Union[Sequence[int], Dict[str, Sequence[int]]]] = None,
    test_times: Optional[Sequence[int]] = None,
    sample_times_absolute: Optional[bool] = False,
    **kwargs,
):
    ax: plt.Axes = kwargs.get("ax")

    test_times = test_times or [0, 0]
    xlabel = kwargs.get("xlabel", _saliva_plot_params.get("xlabel"))
    ylabel = kwargs.get("ylabel", _saliva_plot_params.get("ylabel"))
    xticks = kwargs.get("xticks")
    xaxis_tick_locator = kwargs.get("xaxis_tick_locator")

    data, ylabel = _saliva_plot_sanitize_dicts(data, ylabel, saliva_type)

    _assert_saliva_data_input(data, saliva_type)

    data = data.copy()

    if sample_times is None and "time" not in data.reset_index().columns:
        x = "sample"
        xlabel = "Sample"
    else:
        sample_times = _get_sample_times(data, sample_times, test_times, sample_times_absolute)
        if "time" in data.index.names:
            data.index = data.index.droplevel("time")
        data["time"] = sample_times * int(len(data) / len(sample_times))
        x = "time"

    kwargs.setdefault("hue", "condition" if "condition" in data.index.names else None)
    kwargs.setdefault("style", kwargs.get("hue"))
    kwargs.setdefault("marker", "o")

    if counter == 0 and len(ax.lines) == 0:
        kwargs.setdefault("palette", _get_palette("fau", 2))
    else:
        kwargs.setdefault("palette", _get_palette("tech", 2))
        # the was already something drawn into the axis => we are using the same axis to add another feature
        ax_twin = ax.twinx()
        kwargs.update({"ax": ax_twin, "show_legend": False})

    kwargs.update({"xlabel": xlabel, "ylabel": ylabel})

    lineplot(data=data, x=x, y=saliva_type, **kwargs)

    _saliva_plot_style_xaxis(xticks, xaxis_tick_locator, ax)


def _assert_saliva_data_input(data: pd.DataFrame, saliva_type: str):
    ret = is_saliva_raw_dataframe(data, saliva_type, raise_exception=False)
    ret = ret or is_saliva_mean_se_dataframe(data, raise_exception=False)
    if not ret:
        raise ValidationError("'data' is expected to be either a SalivaRawDataFrame or a SalivaMeanSeDataFrame!")


def _saliva_plot_get_plot_param(param: Union[Dict[str, str], str], key: str):
    p = param[key] if isinstance(param, dict) else param
    return p


def _saliva_plot_style_xaxis(xticks: Sequence[str], xaxis_tick_locator: mticks.Locator, ax: plt.Axes):
    if xticks is not None and xaxis_tick_locator is not None:
        ax.xaxis.set_major_locator(xaxis_tick_locator)
        ax.xaxis.set_ticks(xticks)


def saliva_plot_combine_legend(fig: plt.Figure, ax: plt.Axes, saliva_types: Sequence[str], **kwargs):
    """Combine multiple legends of ``saliva_plot`` into one joint legend outside of plot.

    If data from multiple saliva types are combined into one plot (e.g., by calling
    :func:`~biopsykit.protocols.plotting.saliva_plot` on the same plot twice) then two separate legend are created.
    This function can be used to combine the two legends into one.

    Parameters
    ----------
    fig : :class:`~matplotlib.figure.Figure`
        figure object
    ax : :class:`~matplotlib.axes.Axes`
        axes object
    saliva_types : list
        list of saliva types in plot
    **kwargs
        additional arguments to customize plot, such as:

        * ``legend_loc``: Location of legend. Default: ``upper center``
        * ``legend_size``: Legend size. Default: ``small``
        * ``rect``: Rectangle in normalized figure coordinates into which the whole subplots area
          (including labels) will fit. Used to conveniently place legend outside of figure.

    """
    legend_loc = kwargs.get("legend_loc", "upper center")
    legend_size = kwargs.get("legend_size", "small")
    rect = kwargs.get("rect", (0, 0, 1.0, 0.95))
    labels = [ax.get_legend_handles_labels()[1] for ax in fig.get_axes()]

    if all(len(label) == 1 for label in labels):
        # only one group
        handles = [ax.get_legend_handles_labels()[0] for ax in fig.get_axes()]
        handles = [h[0] for handle in handles for h in handle]
        labels = [_saliva_plot_params.get("legend_title")[b] for b in saliva_types]
        ncol = len(handles)
        fig.legend(
            handles,
            labels,
            loc=legend_loc,
            ncol=ncol,
            prop={"size": legend_size},
        )
    else:
        handles = [ax.get_legend_handles_labels()[0] for ax in fig.get_axes()]
        handles = [h[0] for handle in handles for h in handle]
        labels = [ax.get_legend_handles_labels()[1] for ax in fig.get_axes()]
        labels = [
            "{}: {}".format(_saliva_plot_params.get("legend_title")[b], " - ".join(label))
            for b, label in zip(saliva_types, labels)
        ]
        ncol = len(handles)

        fig.legend(
            list(zip(handles[::2], handles[1::2])),
            labels,
            loc=legend_loc,
            ncol=ncol,
            numpoints=1,
            handler_map={tuple: HandlerTuple(ndivide=None)},
            prop={"size": legend_size},
        )
    ax.legend().remove()
    fig.tight_layout(pad=1.0, rect=rect)


def saliva_feature_boxplot(
    data: SalivaFeatureDataFrame,
    x: str,
    saliva_type: str,
    hue: Optional[str] = None,
    feature: Optional[str] = None,
    stats_kwargs: Optional[Dict[str, Any]] = None,
    **kwargs,
) -> Tuple[plt.Figure, plt.Axes]:
    """Draw a boxplot with significance brackets, specifically designed for saliva features.

    This is a wrapper of :func:`~biopsykit.plotting.feature_boxplot` that can be used to plot saliva features and
    allows to easily add significance brackets that indicate statistical significance.

    .. note::
        The input data is assumed to be in long-format.


    Parameters
    ----------
    data : :obj:`~biopsykit.utils.datatype_helper.SalivaFeatureDataFrame`
        data to plot
    x : str
        column of x axis in ``data``
    saliva_type : str
        type of saliva data to plot
    hue : str, optional
        column name of grouping variable. Default: ``None``
    feature : str, optional
        name of feature to plot or ``None``
    stats_kwargs : dict, optional
        dictionary with arguments for significance brackets
    **kwargs
        additional arguments that are passed to :func:`~biopsykit.plotting.feature_boxplot` and :func:`~seaborn.boxplot`


    Returns
    -------
    fig : :class:`~matplotlib.figure.Figure`
        figure object
    ax : :class:`~matplotlib.axes.Axes`
        axes object


    See Also
    --------
    :func:`~biopsykit.plotting.feature_boxplot`
        plot features as boxplot
    :class:`~biopsykit.stats.stats.StatsPipeline`
        class to create statistical analysis pipelines and get parameter for plotting significance brackets

    """
    is_saliva_feature_dataframe(data, saliva_type)

    if feature is not None:
        if isinstance(feature, str):
            feature = [feature]
        ylabel = _saliva_feature_boxplot_get_ylabels(saliva_type, feature)
        ylabel = [ylabel[f] for f in feature]
        if len(set(ylabel)) == 1:
            kwargs.setdefault("ylabel", ylabel[0])

        if hue is not None:
            xticklabels = list(_saliva_feature_boxplot_get_xticklabels({f: f for f in feature}).values())
            xticklabels = [x[0] for x in xticklabels]
            kwargs.setdefault("xticklabels", xticklabels)

    return feature_boxplot(data=data, x=x, y=saliva_type, stats_kwargs=stats_kwargs, **kwargs)


def saliva_multi_feature_boxplot(
    data: SalivaFeatureDataFrame,
    saliva_type: str,
    features: Union[Sequence[str], Dict[str, Union[str, Sequence[str]]]],
    hue: Optional[str] = None,
    stats_kwargs: Optional[Dict] = None,
    **kwargs,
) -> Tuple[plt.Figure, Iterable[plt.Axes]]:
    """Draw multiple features as boxplots with significance brackets, specifically designed for saliva features.

    This is a wrapper of :func:`~biopsykit.plotting.multi_feature_boxplot` that can be used to plot saliva features and
    allows to easily add significance brackets that indicate statistical significance.

    .. note::
        The input data is assumed to be in long-format.

    Parameters
    ----------
    data : :obj:`~biopsykit.utils.datatype_helper.SalivaFeatureDataFrame`
        data to plot
    saliva_type : str
        type of saliva data to plot
    hue : str, optional
        column name of grouping variable. Default: ``None``
    features : list of str or dict of str
        features to plot. If ``features`` is a list, each entry must correspond to one feature category in the
        index level specified by ``group``. A separate subplot will be created for each feature.
        If similar features (i.e., different `slope` or `AUC` parameters) should be combined into one subplot,
        ``features`` can be provided as dictionary.
        Then, the dict keys specify the feature category (a separate subplot will be created for each category)
        and the dict values specify the feature (or list of features) that are combined into the subplots.
    stats_kwargs : dict, optional
        nested dictionary with arguments for significance brackets.
        See :func:`~biopsykit.plotting.feature_boxplot` for further information


    Returns
    -------
    fig : :class:`~matplotlib.figure.Figure`
        figure object
    axs : list of :class:`matplotlib.axes.Axes`
        list of subplot axes objects


    See Also
    --------
    :func:`~biopsykit.plotting.multi_feature_boxplot`
        plot multiple features as boxplots
    :class:`~biopsykit.stats.stats.StatsPipeline`
        class to create statistical analysis pipelines and get parameter for plotting significance brackets

    """
    x = kwargs.pop("x", "saliva_feature")

    if isinstance(features, str):
        # ensure list
        features = [features]
    if isinstance(features, list):
        features = {f: [f] if isinstance(f, str) else f for f in features}

    kwargs.setdefault("xticklabels", _saliva_feature_boxplot_get_xticklabels(features))
    kwargs.setdefault("ylabels", _saliva_feature_boxplot_get_ylabels(saliva_type, features))

    return multi_feature_boxplot(
        data,
        x=x,
        y=saliva_type,
        group="saliva_feature",
        features=features,
        hue=hue,
        stats_kwargs=stats_kwargs,
        **kwargs,
    )


def _saliva_feature_boxplot_get_xticklabels(features: Dict[str, str]) -> Dict[str, Sequence[str]]:
    xlabel_dict = {}
    for feature in features:
        cols = features[feature]
        if isinstance(cols, str):
            cols = [cols]
        labels = []
        for c in cols:
            if "slope" in c:
                label = _saliva_feature_params["xticklabels"]["slope"].replace("§", re.findall(r"slope(\w+)", c)[0])
            else:
                label = _saliva_feature_params["xticklabels"][c]
            labels.append(label)

        xlabel_dict[feature] = labels
    return xlabel_dict


def _saliva_feature_boxplot_get_ylabels(saliva_type: str, features: Union[str, Sequence[str]]) -> Dict[str, str]:
    ylabels = _saliva_feature_params["ylabel"][saliva_type]
    if isinstance(features, str):
        features = [features]
    for feature in features:
        if "slope" in feature:
            ylabels[feature] = ylabels["slope"]

    return ylabels


def _plot_get_fig_ax(**kwargs):
    ax: plt.Axes = kwargs.get("ax", None)
    if ax is None:
        fig, ax = plt.subplots(figsize=kwargs.get("figsize"))
    else:
        fig = ax.get_figure()
    return fig, ax



"""Module with functions and utility functions to compute questionnaires."""
from biopsykit.questionnaires import utils
from biopsykit.questionnaires.questionnaires import (
    abi,
    abi_ms,
    ads_l,
    asi,
    asku,
    asq,
    asq_mod,
    besaa,
    bfi_10,
    bfi_k,
    bidr,
    brief_cope,
    cesd,
    clq,
    ctq,
    erq,
    eval_clinic,
    fee,
    fkk,
    fscrs,
    ghq,
    hads,
    idq_post_scan,
    idq_pre_scan,
    ie_4,
    kab,
    kkg,
    lsq,
    mbi_gs,
    mbi_gss,
    mdbf,
    meq,
    midi,
    mkhai,
    mlq,
    mves,
    panas,
    pasa,
    peat,
    pfb,
    phq,
    psqi,
    pss,
    purpose_life,
    resilience,
    rmidi,
    rse,
    rsq,
    sci,
    scs,
    sds,
    sop,
    ssgs,
    sss,
    stadi,
    stai_short,
    state_rumination,
    svf_42,
    svf_120,
    swb,
    swls,
    tb,
    tics_l,
    tics_s,
    trait_rumination,
    tsgs,
    type_d,
    wpi,
)

__all__ = [
    "abi",
    "abi_ms",
    "ads_l",
    "asi",
    "asku",
    "asq",
    "asq_mod",
    "bfi_k",
    "bfi_10",
    "bidr",
    "besaa",
    "brief_cope",
    # "ceca",
    "cesd",
    "clq",
    "ctq",
    "erq",
    "eval_clinic",
    "fee",
    "fkk",
    "fscrs",
    "ghq",
    "hads",
    "idq_pre_scan",
    "idq_post_scan",
    "ie_4",
    "kab",
    "kkg",
    "lsq",
    "mbi_gs",
    "mbi_gss",
    "mdbf",
    "meq",
    "midi",
    "mkhai",
    "mlq",
    "mves",
    "panas",
    "pasa",
    "peat",
    "pfb",
    "phq",
    "psqi",
    "pss",
    "purpose_life",
    "resilience",
    "rmidi",
    "rse",
    "rsq",
    "sci",
    "scs",
    "sds",
    "sop",
    "ssgs",
    "sss",
    "swb",
    "stadi",
    "stai_short",
    "state_rumination",
    "svf_42",
    "svf_120",
    "swls",
    "tb",
    "tics_l",
    "tics_s",
    "trait_rumination",
    "tsgs",
    "type_d",
    "utils",
    "wpi",
]



"""Module representing the Trier Social Stress Test (TSST) protocol."""
from typing import Dict, Optional, Union

from biopsykit.protocols import BaseProtocol


class TSST(BaseProtocol):
    """Class representing the Trier Social Stress Test (TSST)."""

    def __init__(
        self, name: Optional[str] = None, structure: Optional[Dict[str, Union[None, Dict[str, int]]]] = None, **kwargs
    ):
        """Class representing the Trier Social Stress Test (TSST).

        The general structure of the TSST can be specified by passing a ``structure`` dict to the constructor.

        Up to three nested structure levels are supported:
            * 1st level: ``study part``: Different parts of the study where the TSST was conducted, such as: "Pre",
              "TSST", and "Post".
            * 2nd level: ``phase``: Different TSST phases that belong to the same *study part*, such as:
              "Preparation", "Talk", "Math" (for study part "TSST") or "Questionnaires", "Rest"
              (for study part "Pre").
            * 3rd level: ``subphase``: Different TSST subphases that belong to the same *phase*.


        Parameters
        ----------
        name : str
            name of protocol or ``None`` to use "TSST" as default name. Default: ``None``
        structure : dict, optional
            nested dictionary specifying the structure of the TSST study.

            Up to three nested structure levels are supported:

            * 1st level: ``study part``: Different parts of the study where the TSST was conducted, such as: "Pre",
              "TSST", and "Post".
            * 2nd level: ``phase``: Different TSST phases that belong to the same *study part*, such as:
              "Preparation", "Talk", "Math" (for study part "TSST") or "Questionnaires", "Rest"
              (for study part "Pre").
            * 3rd level: ``subphase``: Different TSST subphases that belong to the same *phase*.

            If a study part has no division into finer phases (or a phase has no division into finer subphases) the
            dictionary value can be set to ``None``.
        **kwargs
            additional parameters to be passed to ``MIST`` and its superclass, ``BaseProtocol``, such as:

            * ``saliva_plot_params``: dictionary with parameters to style
              :meth:`~biopsykit.protocols.BaseProtocol.saliva_plot`
            * ``hr_mean_plot_params``: dictionary with parameters to style
              :meth:`~biopsykit.protocols.BaseProtocol.hr_mean_plot`
            * ``hr_ensemble_plot_params``: dictionary with parameters to style
              :meth:`~biopsykit.protocols.BaseProtocol.hr_ensemble_plot`


        Examples
        --------
        >>> from biopsykit.protocols import TSST
        >>> # Example: TSST study consisting of three parts. Only the TSST part consists of different phases
        >>> structure = {
        >>>     "Pre": None,
        >>>     "TSST": {
        >>>         "Preparation": 300,
        >>>         "Talk": 300,
        >>>         "Math": 300
        >>>     },
        >>>     "Post": None
        >>> }
        >>> TSST(name="TSST", structure=structure)


        References
        ----------
        Kirschbaum, C., Pirke, K.-M., & Hellhammer, D. H. (1993). The “Trier Social Stress Test” - A Tool for
        Investigating Psychobiological Stress Responses in a Laboratory Setting. *Neuropsychobiology*, 28, 76-81.
        https://doi.org/10.1159/000119004

        """
        if name is None:
            name = "TSST"
        if structure is None:
            structure = {"Part1": None, "TSST": {"Preparation": 300, "Talk": 300, "Math": 300}, "Part2": None}

        test_times = kwargs.pop("test_times", [0, 15])
        saliva_plot_params = {"test_title": "TSST", "xlabel": "Time relative to TSST start [min]"}
        saliva_plot_params.update(kwargs.pop("saliva_plot_params", {}))

        kwargs.update({"saliva_plot_params": saliva_plot_params})

        super().__init__(name=name, structure=structure, test_times=test_times, **kwargs)



"""Module with a set of functions useful for computing."""
from typing import Union

import numpy as np
import pandas as pd


def se(data: Union[np.ndarray, pd.Series]) -> Union[np.ndarray, pd.Series]:
    """Compute standard error (SE).

    .. note::
        For computing the standard error the *corrected* sample standard deviation (``ddof = 1``) is used

    Parameters
    ----------
    data : array_like
        input data

    Returns
    -------
    array_like
        standard error of data

    """
    return np.std(data, ddof=1) / np.sqrt(len(data))


def mean_se(data: pd.DataFrame) -> pd.DataFrame:
    """Compute mean and standard error from a dataframe.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        input data

    Returns
    -------
    :class:`~pandas.DataFrame`
        mean and standard error of data

    """
    return data.agg([np.mean, se])



"""Module providing functions to load and save logs from the *CARWatch* app."""
import json
import re
import warnings
import zipfile
from pathlib import Path
from typing import TYPE_CHECKING, Dict, Optional, Sequence, Union

import pandas as pd
from biopsykit.utils._datatype_validation_helper import _assert_file_extension
from biopsykit.utils._types import path_t
from biopsykit.utils.time import tz, utc
from tqdm.auto import tqdm

if TYPE_CHECKING:
    from biopsykit.carwatch_logs import LogData

LOG_FILENAME_PATTERN = "logs_(.*?)"


def load_logs_all_subjects(
    base_folder: path_t,
    has_subject_folders: Optional[bool] = True,
    log_filename_pattern: Optional[str] = None,
    return_df: Optional[bool] = True,
) -> Union[pd.DataFrame, Dict[str, pd.DataFrame]]:
    """Load log files from all subjects in a folder.

    This function iterates through the base folder and looks for subfolders
    (if ``has_subject_folders`` is ``True``), or for .csv files or .zip files matching the log file name pattern.

    Files from all subjects are then loaded and returned as one :class:`~pandas.DataFrame`
    (if ``return_df`` is ``True``) or a dictionary (if ``return_df`` is ``False``).


    Parameters
    ----------
    base_folder : str or :class:`~pathlib.Path`
        path to base folder containing log files
    has_subject_folders : boolean, optional
        ``True`` if log files are stored in subfolders per subject, ``False`` if they are all stored in one
        top-level folder
    log_filename_pattern : str, optional
        file name pattern of log files as regex string or ``None`` if files have default filename
        pattern: "logs_(.*?)". A custom filename pattern needs to contain a capture group to extract the subject ID
    return_df : bool, optional
        ``True`` to return data from all subjects combined as one dataframe, ``False`` to return a dictionary with
        data per subject. Default: ``True``


    Returns
    -------
    :class:`~pandas.DataFrame` or dict
        dataframe with log data for all subjects (if ``return_df`` is ``True``).
        or dictionary with log data per subject

    """
    # ensure pathlib
    base_folder = Path(base_folder)

    if has_subject_folders:
        folder_list = [p for p in sorted(base_folder.glob("*")) if p.is_dir() and not p.name.startswith(".")]
        dict_log_files = _load_log_file_folder(folder_list)
    else:
        # first, look for available csv files
        file_list = sorted(base_folder.glob("*.csv"))
        if len(file_list) > 0:
            dict_log_files = _load_log_file_list_csv(file_list, log_filename_pattern)
        else:
            # fallback: look for zip files
            file_list = sorted(base_folder.glob("*.zip"))
            dict_log_files = _load_log_file_zip(file_list, log_filename_pattern)

    if return_df:
        return pd.concat(dict_log_files, names=["subject_id"])
    return dict_log_files


def _load_log_file_folder(folder_list: Sequence[Path]):
    dict_log_files = {}
    for folder in tqdm(folder_list):
        subject_id = folder.name
        dict_log_files[subject_id] = load_log_one_subject(folder)

    return dict_log_files


def _load_log_file_list_csv(file_list: Sequence[Path], log_filename_pattern: str):
    dict_log_files = {}
    if log_filename_pattern is None:
        log_filename_pattern = LOG_FILENAME_PATTERN + ".csv"
    for file in tqdm(file_list):
        subject_id = re.search(log_filename_pattern, file.name).group(1)
        df = pd.read_csv(file, sep=";")
        df["time"] = pd.to_datetime(df["time"])
        df = df.set_index("time")
        dict_log_files[subject_id] = df
    return dict_log_files


def _load_log_file_zip(file_list: Sequence[Path], log_filename_pattern: str) -> Dict[str, pd.DataFrame]:
    dict_log_files = {}
    if log_filename_pattern is None:
        log_filename_pattern = LOG_FILENAME_PATTERN + ".zip"
    for file in tqdm(file_list):
        subject_id = re.search(log_filename_pattern, file.name).group(1)
        dict_log_files[subject_id] = load_log_one_subject(file)

    return dict_log_files


def load_log_one_subject(
    path: path_t,
    log_filename_pattern: Optional[str] = None,
    overwrite_unzipped_logs: Optional[bool] = False,
) -> pd.DataFrame:
    """Load log files from one subject.

    Parameters
    ----------
    path : :class:`~pathlib.Path` or str
        path to folder containing log files from subject or path to log file from subject
    log_filename_pattern : str, optional
        file name pattern of log files as regex string or ``None`` if file has default filename
        pattern: "logs_(.*?)". A custom filename pattern needs to contain a capture group to extract the subject ID
    overwrite_unzipped_logs : bool, optional
        ``True`` to overwrite already unzipped log files, ``False`` to not overwrite.
        Only relevant if log files are provided as zip files. Default: ``False``

    Returns
    -------
    :class:`~pandas.DataFrame`
        dataframe with log data for one subject

    """
    if log_filename_pattern is None:
        log_filename_pattern = LOG_FILENAME_PATTERN + ".zip"

    path = Path(path)

    if path.is_dir():
        # TODO add error messages if no log files are found (e.g. wrong folder path etc.)
        return log_folder_to_dataframe(path)

    if path.is_file():
        _assert_file_extension(path, [".zip", ".csv"])
    if path.suffix == ".zip":
        with zipfile.ZipFile(path, "r") as zip_ref:
            export_folder = path.parent.joinpath(re.search(log_filename_pattern, path.name).group(1))
            export_folder.mkdir(exist_ok=True)
            if overwrite_unzipped_logs or len(list(export_folder.glob("*"))) == 0:
                zip_ref.extractall(export_folder)
                # call recursively with newly exported folder
                return log_folder_to_dataframe(export_folder)
            # folder not empty => inform user and load folder
            warnings.warn(
                f"Folder {export_folder.name} already contains log files which will be loaded. "
                f"Set `overwrite_logs_unzip = True` to overwrite log files."
            )
            return log_folder_to_dataframe(export_folder)
    return _load_log_file_csv(path)


def log_folder_to_dataframe(folder_path: path_t) -> pd.DataFrame:
    """Load log data from folder of one subject and return it as dataframe.

    Parameters
    ----------
    folder_path : :class:`~pathlib.Path` or str
        path to folder containing log files from subject


    Returns
    -------
    :class:`~pandas.DataFrame`
        dataframe with log data for one subject

    """
    file_list = sorted(folder_path.glob("*.csv"))
    return pd.concat([_load_log_file_csv(file) for file in file_list])


def _load_log_file_csv(file_path: path_t) -> pd.DataFrame:
    df = pd.read_csv(file_path, sep=";", header=None, names=["time", "action", "extras"])

    df["time"] = pd.to_datetime(df["time"], unit="ms")
    df = df.set_index("time")
    df.index = df.index.tz_localize(utc).tz_convert(tz)
    df = df.sort_index()
    df = df.apply(_parse_date, axis=1)
    return df


def save_log_data(
    log_data: Union[pd.DataFrame, "LogData"],
    path: path_t,
    subject_id: Optional[str] = None,
    overwrite: Optional[bool] = False,
    show_skipped: Optional[bool] = False,
):
    """Save log data for a single subject or for all subjects at once.

    The behavior of this function depends on the input passed to ``log_data``:

    * If ``log_data`` is a :class:`~pandas.DataFrame` with a :class:`~pandas.MultiIndex` it is assumed that
      the dataframe contains data from multiple subjects and will be exported accordingly as one combined csv file.
    * If ``log_data`` is a :class:`~pandas.DataFrame` without :class:`~pandas.MultiIndex` it is assumed that the
      dataframe only contains data from one single subject and will be exported accordingly as csv file.

    Parameters
    ----------
    log_data : :class:`~pandas.DataFrame` or :class:`~biopsykit.carwatch_logs.log_data.LogData`
        log data to save
    path : :class:`~pathlib.Path` or str
        path for export. The expected format of ``path`` depends on ``log_data``:

        * If ``log_data`` is log data from a single subject ``path`` needs to specify a **folder**.
          The log data will then be exported to "path/logs_<subject_id>.csv".
        * If ``log_data`` is log data from multiple subjects ``path`` needs to specify a **file**.
          The combined log data of all subjects will then be exported to "path".

    subject_id : str, optional
        subject ID or ``None`` to get subject ID from ``LogData`` object
    overwrite : bool, optional
        ``True`` to overwrite file if it already exists, ``False`` otherwise. Default: ``False``
    show_skipped : bool, optional
        ``True`` to print message if log data was already exported and will be skipped, ``False`` otherwise.
        Default: ``False``

    """
    from biopsykit.carwatch_logs import LogData  # pylint: disable=import-outside-toplevel

    if isinstance(log_data, pd.DataFrame):
        if isinstance(log_data.index, pd.MultiIndex):
            # dataframe has a multiindex => it's a combined dataframe for all subjects
            log_data.to_csv(path, sep=";")
            return
        log_data = LogData(log_data)

    if subject_id is None:
        subject_id = log_data.subject_id

    export_path = path.joinpath(f"logs_{subject_id}.csv")
    if not export_path.exists() or overwrite:
        log_data.data.to_csv(export_path, sep=";")
    elif show_skipped:
        print(f"Skipping subject {subject_id}. Already exported.")


def _parse_date(row: pd.Series) -> pd.Series:
    """Parse date in "timestamp" and "timestamp_hidden" columns of a pandas series.

    Parameters
    ----------
    row : :class:`~pandas.Series`
        one row of a dataframe

    Returns
    -------
    :class:`~pandas.Series`
        series with parsed date

    """
    json_extra = json.loads(row["extras"])
    row_cpy = row.copy()
    keys = ["timestamp", "timestamp_hidden"]
    for key in keys:
        if key in json_extra:
            # convert to datetime and localize
            time = utc.localize(pd.to_datetime(json_extra[key], unit="ms"))
            # convert to correct time zone
            time = time.astimezone(tz)
            json_extra[key] = str(time)

    # fix wrong key for saliva_id in "alarm_ring" action (old app versions)
    if "extra_saliva_id" in json_extra:
        json_extra["saliva_id"] = json_extra["extra_saliva_id"]
        del json_extra["extra_saliva_id"]
    row_cpy["extras"] = json.dumps(json_extra)
    return row_cpy



"""Base class for implementing signal processing pipelines."""
from typing import Dict, Optional, Sequence, Union

import pandas as pd
from biopsykit.utils._datatype_validation_helper import _assert_is_dtype
from biopsykit.utils.data_processing import split_data


class _BaseProcessor:
    """Base class for implementing signal processing pipelines in BioPsyKit."""

    def __init__(
        self,
        data: Union[pd.DataFrame, Dict[str, pd.DataFrame]],
        sampling_rate: Optional[float] = None,
        time_intervals: Optional[Union[pd.Series, Dict[str, Sequence[str]]]] = None,
        include_start: Optional[bool] = False,
    ):
        self.sampling_rate: float = sampling_rate
        """Sampling rate of recorded data."""

        if isinstance(data, dict):
            for _, df in data.items():
                _assert_is_dtype(df, pd.DataFrame)
            data_dict = data
        else:
            _assert_is_dtype(data, pd.DataFrame)
            if time_intervals is not None:
                # split data into subphases if time_intervals are passed
                data_dict = split_data(
                    data=data,
                    time_intervals=time_intervals,
                    include_start=include_start,
                )
            else:
                data_dict = {"Data": data}

        self.data: Dict[str, pd.DataFrame] = data_dict
        """Dictionary with raw data, split into different phases.

        Each dataframe is expected to be a :class:`~pandas.DataFrame`.
        """

    @property
    def phases(self) -> Sequence[str]:
        """List of phases.

        Returns
        -------
        list
            phase names

        """
        return list(self.data.keys())



"""Module wrapping biopsykit.io.biomarker including only I/O functions for saliva data."""
from pathlib import Path
from typing import Dict, Optional, Sequence, Tuple, Union

import pandas as pd
from biopsykit.io import biomarker

__all__ = ["load_saliva_plate", "save_saliva", "load_saliva_wide_format"]

from biopsykit.utils._types import path_t
from biopsykit.utils.datatype_helper import SalivaRawDataFrame, SubjectConditionDataFrame


def load_saliva_plate(
    file_path: path_t,
    saliva_type: str,
    sample_id_col: Optional[str] = None,
    data_col: Optional[str] = None,
    id_col_names: Optional[Sequence[str]] = None,
    regex_str: Optional[str] = None,
    sample_times: Optional[Sequence[int]] = None,
    condition_list: Optional[Union[Sequence, Dict[str, Sequence], pd.Index]] = None,
    **kwargs,
) -> SalivaRawDataFrame:
    r"""Read saliva from an Excel sheet in 'plate' format.

    This function automatically extracts identifier like subject, day and sample IDs from the saliva sample names.
    To extract them, a regular expression string can be passed via ``regex_str``.

    Here are some examples on how sample identifiers might look like and what the corresponding
    ``regex_str`` would output:

        * "Vp01 S1"
          => ``r"(Vp\d+) (S\d)"`` (this is the default pattern, you can also just set ``regex_str`` to ``None``)
          => data ``[Vp01, S1]`` in two columns: ``subject``, ``sample``
          (unless column names are explicitly specified in ``data_col_names``)
        * "Vp01 T1 S1" ... "Vp01 T1 S5" (only *numeric* characters in day/sample)
          => ``r"(Vp\d+) (T\d) (S\d)"``
          => three columns: ``subject``, ``sample`` with data ``[Vp01, T1, S1]``
          (unless column names are explicitly specified in ``data_col_names``)
        * "Vp01 T1 S1" ... "Vp01 T1 SA" (also *letter* characters in day/sample)
          => ``r"(Vp\d+) (T\w) (S\w)"``
          => three columns: ``subject``, ``sample`` with data ``[Vp01, T1, S1]``
          (unless column names are explicitly specified in ``data_col_names``)

    If you **don't** want to extract the 'S' or 'T' prefixes in saliva or day IDs, respectively,
    you have to move it **out** of the capture group in the ``regex_str`` (round brackets), like this:
    ``(S\d)`` (would give ``S1``, ``S2``, ...)
    => ``S(\d)`` (would give ``1``, ``2``, ...)


    Parameters
    ----------
    file_path: :class:`~pathlib.Path` or str
        path to the Excel sheet in 'plate' format containing saliva data
    saliva_type: str
        saliva type to load from file
    sample_id_col: str, optional
        column name of the Excel sheet containing the sample ID. Default: "sample ID"
    data_col: str, optional
        column name of the Excel sheet containing saliva data to be analyzed.
        Default: Select default column name based on ``biomarker_type``, e.g. ``cortisol`` => ``cortisol (nmol/l)``
    id_col_names: list of str, optional
        names of the extracted ID column names. ``None`` to use the default column names (['subject', 'day', 'sample'])
    regex_str: str, optional
        regular expression to extract subject ID, day ID and sample ID from the sample identifier.
        ``None`` to use default regex string (``r"(Vp\d+) (S\d)"``)
    sample_times: list of int, optional
        times at which saliva samples were collected
    condition_list: 1d-array, optional
        list of conditions which subjects were assigned to
    **kwargs
        Additional parameters that are passed to :func:`pandas.read_excel`

    Returns
    -------
    data : :class:`~biopsykit.utils.datatype_helper.SalivaRawDataFrame`
        saliva data in `SalivaRawDataFrame` format

    Raises
    ------
    :exc:`~biopsykit.utils.exceptions.FileExtensionError`
        if file is no Excel file (.xls or .xlsx)
    ValueError
        if any saliva sample can not be converted into a float (e.g. because there was text in one of the columns)
    :exc:`~biopsykit.utils.exceptions.ValidationError`
        if imported data can not be parsed to a SalivaRawDataFrame

    """
    return biomarker.load_saliva_plate(
        file_path, saliva_type, sample_id_col, data_col, id_col_names, regex_str, sample_times, condition_list, **kwargs
    )


def save_saliva(
    file_path: path_t,
    data: SalivaRawDataFrame,
    saliva_type: Optional[str] = "cortisol",
    as_wide_format: Optional[bool] = False,
):
    """Save saliva data to csv file.

    Parameters
    ----------
    file_path: :class:`~pathlib.Path` or str
        file path to export. Must be a csv or an Excel file
    data : :class:`~biopsykit.utils.datatype_helper.SalivaRawDataFrame`
        saliva data in `SalivaRawDataFrame` format
    saliva_type : str
        type of saliva data in the dataframe
    as_wide_format : bool, optional
        ``True`` to save data in wide format (and flatten all index levels), ``False`` to save data in long-format.
        Default: ``False``

    Raises
    ------
    :exc:`~biopsykit.utils.exceptions.ValidationError`
        if ``data`` is not a SalivaRawDataFrame
    :exc:`~biopsykit.utils.exceptions.FileExtensionError`
        if ``file_path`` is not a csv or Excel file

    """
    biomarker.save_saliva(file_path, data, saliva_type, as_wide_format)


def load_saliva_wide_format(
    file_path: path_t,
    saliva_type: str,
    subject_col: Optional[str] = None,
    condition_col: Optional[str] = None,
    additional_index_cols: Optional[Union[str, Sequence[str]]] = None,
    sample_times: Optional[Sequence[int]] = None,
    **kwargs,
) -> SalivaRawDataFrame:
    """Load saliva data that is in wide-format from csv file.

    It will return a `SalivaRawDataFrame`, a long-format dataframe that complies with BioPsyKit's naming convention,
    i.e., the subject ID index will be named ``subject``, the sample index will be names ``sample``,
    and the value column will be named after the saliva biomarker type.

    Parameters
    ----------
    file_path: :class:`~pathlib.Path` or str
        path to file
    saliva_type: str
        saliva type to load from file. Example: ``cortisol``
    subject_col: str, optional
        name of column containing subject IDs or ``None`` to use the default column name ``subject``.
        According to BioPsyKit's convention, the subject ID column is expected to have the name ``subject``.
        If the subject ID column in the file has another name, the column will be renamed in the dataframe
        returned by this function. Default: ``None``
    condition_col : str, optional
        name of the column containing condition assignments or ``None`` if no conditions are present.
        According to BioPsyKit's convention, the condition column is expected to have the name ``condition``.
        If the condition column in the file has another name, the column will be renamed in the dataframe
        returned by this function. Default: ``None``
    additional_index_cols : str or list of str, optional
        additional index levels to be added to the dataframe, e.g., "day" index. Can either be a string or a list
        strings to indicate column name(s) that should be used as index level(s),
        or ``None`` for no additional index levels. Default: ``None``
    sample_times: list of int, optional
        times at which saliva samples were collected or ``None`` if no sample times should be specified.
        Default: ``None``
    **kwargs
        Additional parameters that are passed to :func:`pandas.read_csv` or :func:`pandas.read_excel`

    Returns
    -------
    data : :class:`~biopsykit.utils.datatype_helper.SalivaRawDataFrame`
        saliva data in `SalivaRawDataFrame` format

    Raises
    ------
    :exc:`~biopsykit.utils.exceptions.FileExtensionError`
        if file is no csv or Excel file

    """
    return biomarker.load_saliva_wide_format(
        file_path, saliva_type, subject_col, condition_col, additional_index_cols, sample_times, **kwargs
    )


def _get_index_cols(condition_col: str, index_cols: Sequence[str], additional_index_cols: Sequence[str]):
    return biomarker._get_index_cols(condition_col, index_cols, additional_index_cols)


def _read_dataframe(file_path: Path, **kwargs):
    return biomarker._read_dataframe(file_path, **kwargs)


def _check_num_samples(num_samples: int, num_subjects: int):
    """Check that number of imported samples is the same for all subjects.

    Parameters
    ----------
    num_samples : int
        total number of saliva samples in the current dataframe
    num_subjects : int
        total number of subjects in the current dataframe

    Raises
    ------
    ValueError
        if number of samples is not equal for all subjects

    """
    biomarker._check_num_samples(num_samples, num_subjects)


def _check_sample_times(num_samples: int, num_subjects: int, sample_times: Sequence[int]):
    """Check that sample times have the correct number of samples and are monotonously increasing.

    Parameters
    ----------
    num_samples : int
        total number of saliva samples in the current dataframe
    num_subjects : int
        total number of subjects in the current dataframe
    sample_times : array-like
        list of sample times

    Raises
    ------
    ValueError
        if values in ``sample_times`` are not monotonously increasing or
        if number of saliva times does not match the number of saliva samples per subject

    """
    biomarker._check_sample_times(num_samples, num_subjects, sample_times)


def _parse_condition_list(
    data: pd.DataFrame, condition_list: Union[Sequence, Dict[str, Sequence], pd.Index]
) -> SubjectConditionDataFrame:
    return biomarker._parse_condition_list(data, condition_list)


def _apply_condition_list(
    data: pd.DataFrame,
    condition_list: Optional[Union[Sequence, Dict[str, Sequence], pd.Index]] = None,
):
    return biomarker._apply_condition_list(data, condition_list)


def _get_id_columns(id_col_names: Sequence[str], extracted_cols: pd.DataFrame):

    return biomarker._get_id_columns(id_col_names, extracted_cols)


def _get_condition_col(data: pd.DataFrame, condition_col: str) -> Tuple[pd.DataFrame, str]:
    return biomarker._get_condition_col(data, condition_col)
